import os
import sys
from warnings import warn

from rdflib import Namespace
from rdflib.plugins.serializers.turtle import TurtleSerializer

""" 
URIs, Local Names and Namespaces 

Physical layout:
    biolinkml/
        |
        +------ meta.yaml
        |
        +------ context.jsonld  (Master context for metamodel)
        |
        +------ meta.jsonld     (Jsonld representation of metamodel)
        |
        +------ meta.shex       (ShExC representation of metamodel)
        |
        +------ meta.shexj      (ShExJ representation of metamodel)
        |
        +------ meta.ttl        (Metamodel in RDF)
        |
        +------ meta.owl
        |
        +------ includes/
        |          |
        |          +--- mappings.context.jsonld
        |          |
        |          +--- mappings.jsonld
        |          |
        |          +--- mappings.py
        |          |
        |          +--- mappings.yaml
        |          |
        |          +--- types.context.jsonld
        |          |
        |          +--- types.jsonld
        |          |
        |          +--- types.py
        |          |
        |          +--- types.yaml
        |
        +------ biolinkml/
        |          |
        |           +--- meta.py
        |
        +------ meta_mappings_docs/
                  |
                  +--- abstract.md
                  |
                  +--- Element.md
                  |
                  +---    ...
                  |
                  +--- types/
                         |
                         +--- boolean.md


URI Maps:
    # Access to the root directory -- the whole project
    https://w3id.org/biolink/biolinkml        --> biolinkml
    
    # Access to the entire metamodel in various formats
    https://w3id.org/biolink/biolinkml/meta   --> biolinkml/meta   (.yaml, .shex, .ttl, .owl) -- conneg
    
    # Access to documentation on metamodel components
    https://w3id.org/biolink/biolinkml/meta/  --> biolink/meta_mappings_docs/
    
    # Access to the entire types model in various formats
    https://w3id.org/biolink/biolinkml/types  --> biolink/includes/types (.yaml, .shex, .ttl, .owl) -- conneg
    
    # Access to documentation on type components
    https://w3id.org/biolink/biolinkml/types/ --> biolink/meta_mappings_docs/types/



"""

METAMODEL_FILE_NAME = 'meta.yaml'
METAMODEL_LDCONTEXT_NAME = 'context.jsonld'
METAMODEL_SHEXC_NAME = 'meta.shex'
METAMODEL_SHEXJ_NAME = 'meta.shexj'
METAMODEL_RDF_NAME = 'meta.ttl'
METAMODEL_JSONLD_NAME = 'meta.jsonld'
TYPES_FILE_NAME = 'types.yaml'
TYPES_LDCONTEXT_NAME = 'types.context.jsonld'
TYPES_JSONLD_NAME = 'types.jsonld'
MAPPING_FILE_NAME = 'mappings.yaml'
MAPPING_LDCONTEXT_NAME = 'types.context.jsonld'
MAPPING_JSONLD_NAME = 'types.jsonld'


MODULE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
INCLUDES_DIR = os.path.join(MODULE_DIR, 'includes')

# Local location of yaml files
LOCAL_METAMODEL_YAML_FILE = os.path.join(MODULE_DIR, METAMODEL_FILE_NAME)
LOCAL_TYPES_YAML_FILE = os.path.join(INCLUDES_DIR, TYPES_FILE_NAME)
LOCAL_MAPPING_YAML_FILE = os.path.join(INCLUDES_DIR, MAPPING_FILE_NAME)

# Local location of jsonld and context.jsonld files
LOCAL_METAMODEL_LDCONTEXT_FILE = os.path.join(MODULE_DIR, METAMODEL_LDCONTEXT_NAME)
LOCAL_METAMODEL_JSONLD_FILE = os.path.join(MODULE_DIR, METAMODEL_JSONLD_NAME)
LOCAL_TYPES_LDCONTEXT_FILE = os.path.join(INCLUDES_DIR, TYPES_LDCONTEXT_NAME)
LOCAL_TYPES_JSONLD_FILE = os.path.join(INCLUDES_DIR, TYPES_JSONLD_NAME)
LOCAL_MAPPING_LDCONTEXT_FILE = os.path.join(INCLUDES_DIR, MAPPING_LDCONTEXT_NAME)
LOCAL_MAPPING_JSONLD_FILE = os.path.join(INCLUDES_DIR, MAPPING_JSONLD_NAME)

# Local location of metamodel shex file
LOCAL_SHEXJ_FILE_NAME = os.path.join(MODULE_DIR, METAMODEL_SHEXJ_NAME)
LOCAL_SHEXC_FILE_NAME = os.path.join(MODULE_DIR, METAMODEL_SHEXC_NAME)

# Local location of the metamodel rdf file
LOCAL_RDF_FILE_NAME = os.path.join(MODULE_DIR, METAMODEL_RDF_NAME)

# Base URI for all things meta
META_BASE_URI = 'https://w3id.org/biolink/biolinkml/'

# URI for the entire metamodel itself.
METAMODEL_NAME = 'metamodel'
METAMODEL_URI = META_BASE_URI + 'meta'
METATYPE_NAME = 'types'
METATYPE_URI = META_BASE_URI + METATYPE_NAME
METAMAPPING_NAME = 'mappings'
METAMAPPING_URI = META_BASE_URI + METAMAPPING_NAME

# Preferred local name for metamodel elements
METAMODEL_NAMESPACE_NAME = "meta"
METATYPE_NAMESPACE_NAME = 'metatype'
METAMAPPING_NAMESPACE_NAME = 'meta'

# Namespace for metamodel elements
METAMODEL_NAMESPACE = Namespace(METAMODEL_URI + '/')
METATYPE_NAMESPACE = Namespace(META_BASE_URI + 'type/')
METAMAPPING_NAMESPACE = METAMODEL_NAMESPACE

# Metamodel Context URI
METAMODEL_CONTEXT_URI = META_BASE_URI + METAMODEL_LDCONTEXT_NAME

# Metamodel ShEx URI
METAMODEL_SHEXJ_URI = META_BASE_URI + 'meta.shexj'
METAMODEL_SHEXC_URI = META_BASE_URI + 'meta.shexc'

# Metamodel YAML file
METAMODEL_YAML_URI = META_BASE_URI + 'meta.yaml'

# Biolink model file -- this needs a more official fix
BIOLINK_MODEL_URI = "https://w3id.org/biolink/biolink-model"
BIOLINK_MODEL_PYTHON_LOC = "biolink.model"

# Make sure we've got the proper rdflib-jsonld
# from rdflib_jsonld import __version__ as rdflib_jsonld_version
# if rdflib_jsonld_version != '0.5.1':
#     warn("https://github.com/hsolbrig/rdflib-jsonld must be installed for non-standard (e.g. CHEBI) prefix generation")
#     warn('RUN: \'pip install "git+git://github.com/hsolbrig/rdflib-jsonld@master#egg=rdflib-jsonld" --upgrade\'')

TurtleSerializer.roundtrip_prefixes = ['']

if sys.version_info < (3, 7, 6):
    warn(f"Some URL processing will fail with python 3.7.5 or earlier.  Current version: {sys.version_info}")


import re
from typing import Callable, Optional, Text, List, Tuple, Match

from biolinkml.meta import ClassDefinition, SlotDefinition
from biolinkml.utils.formatutils import sfx
from biolinkml.utils.schemaloader import SchemaLoader


def strval(txt: str) -> str:
    txt = str(txt).replace('"', '\\"')
    return f'"{txt}"'


def default_uri_for(loader: SchemaLoader) -> str:
    dflt = loader.schema.default_prefix if loader.schema.default_prefix else sfx(loader.schema.id)
    return sfx(loader.namespaces.uri_for(dflt))


def default_curie_or_uri(loader: SchemaLoader) -> str:
    dflt = loader.schema.default_prefix if loader.schema.default_prefix else sfx(loader.schema.id)
    if ':/' in dflt:
        prefix = loader.namespaces.prefix_for(loader.schema.default_prefix)
        if prefix:
            dflt = prefix
    return dflt


def curie_for(loader: SchemaLoader, is_class: bool) -> Optional[str]:
    """ Return the Curie for the schema in loader.  Return None if there is no curie form """
    prefix = default_curie_or_uri(loader)
    suffix = "camelcase(self.name)" if is_class else "underscore(self.alias if self.alias else self.name)"
    if ':/' not in prefix:
        return '"' + prefix + ':' + '" + ' + suffix
    else:
        pn = loader.namespaces.curie_for(prefix, default_ok=False)
        return ('"' + pn + '" + ' + suffix) if pn else None


def uri_for(s: str, loader: SchemaLoader) -> str:
    uri = str(loader.namespaces.uri_for(s))
    return loader.namespaces.curie_for(uri, True, True) or strval(uri)


def default_ns_for(loader: SchemaLoader, cls: ClassDefinition) -> str:
    """ Return code to produce the default namespace for the supplied class """
    # TODO: figure out how to mark a slot as a namespace
    return "sfx(str(self.id))" if "id" in cls.slots else "None"
    # cls_id = None
    # for slotname in cls.slots:
    #     slot = loader.schema.slots[slotname]
    #     if slot.identifier:
    #         cls_id = slotname
    # return f"sfx(str(self.{cls_id}))" if cls_id else "None"

# Library of named default values -- this is here to prevent code injection
# Contents: Match text (as re),
#           flag that indicates whether we're generating a default value expression or postinig code
#           Function that takes the match string, SchemaLoader, ClassDefinition, and SlotDefinition and returns the
#           appropriate string
default_library: List[Tuple[Text, bool, Callable[[Match[str], SchemaLoader, ClassDefinition, SlotDefinition], str]]] = [
    (r"[Tt]rue", False, lambda _, __, ___, ____: "True"),
    (r"[Ff]alse", False, lambda _, __, ___, ____: "False"),
    (r"int\((-?[1-9][0-9]*)\)", False, lambda m, __, ___, ____: int(m[1])),
    # TODO: We have to make the real URI available before any of these can work
    # ("class_uri", True, lambda _, loader, ___, ____: f'"{default_uri_for(loader)}" + camelcase(self.name)'),
    # ("slot_uri", True, lambda _, loader, ___, ____: f'"{default_uri_for(loader)}" + underscore(self.alias if self.alias else self.name)'),
    # ("class_curie", True, lambda _, loader, ___, ____: curie_for(loader, True)),
    # ("slot_curie", True, lambda _, loader, ___, ____: curie_for(loader, False)),
    ("class_uri", True, lambda _, loader, ___, ____: "None"),
    ("slot_uri", True, lambda _, loader, ___, ____: "None"),
    ("class_curie", True, lambda _, loader, ___, ____: "None"),
    ("slot_curie", True, lambda _, loader, ___, ____: "None"),
    # TODO: If you assign a range in the consructor, mergeutils has no way of knowing whether the range
    #       was overridden or just defaulted.  We need to let the old code continue to work until we get
    #       this bit resolved
    # ("default_range", False, lambda _, loader, __, ____: f"ElementName({strval(loader.schema.default_range)})"),
    ("default_range", False, lambda _, __, ___, ____: "None"),
    ("bnode", False, lambda _, __, ___, ____: "bnode()"),
    (r"string\((.*)\)", False, lambda m, __, ___, ____: strval(m[1])),
    (r"uri\((.*)\)", False, lambda m, loader, _, __: uri_for(m[1], loader)),
    ("default_ns", True, lambda _, loader, cls, ____: default_ns_for(loader, cls))
]


def isabsent_match(txt: Text) -> \
        Optional[Tuple[Match[str], bool, Callable[[Match[str], SchemaLoader, ClassDefinition, SlotDefinition], str]]]:
    txt = str(txt)
    for pattern, postinit, f in default_library:
        m = re.match(pattern + '$', txt)
        if m:
            return m, postinit, f


def ifabsent_value_declaration(txt: Text, loader, cls, slot) -> Optional[str]:
    m, postinit, f = isabsent_match(txt)
    if m and not postinit:
        return f(m, loader, cls, slot)


def ifabsent_postinit_declaration(txt: Text, loader, cls, slot) -> Optional[str]:
    m, postinit, f = isabsent_match(txt)
    if m and postinit:
        return f(m, loader, cls, slot)


import abc
import logging
from contextlib import redirect_stdout
from io import StringIO
from typing import List, Set, Union, TextIO, Optional, cast, Callable

import click
from click import Command, Argument, Option

from biolinkml.meta import SchemaDefinition, ClassDefinition, SlotDefinition, ClassDefinitionName, \
    TypeDefinition, Element, SlotDefinitionName, TypeDefinitionName, PrefixPrefixPrefix, ElementName, \
    SubsetDefinition, SubsetDefinitionName
from biolinkml.utils.formatutils import camelcase, underscore
from biolinkml.utils.mergeutils import alias_root
from biolinkml.utils.schemaloader import SchemaLoader
from biolinkml.utils.typereferences import References

DEFAULT_LOG_LEVEL: str = "WARNING"
DEFAULT_LOG_LEVEL_INT: str = logging.WARNING


class Generator(metaclass=abc.ABCMeta):
    generatorname: str = None                   # Set to os.path.basename(__file__)
    generatorversion: str = None                # Generator version identifier
    valid_formats: List[str] = []               # Allowed formats
    base_dir: str = None                        # Base directory of schema

    visit_all_class_slots: bool = False         # False means only visit own slots, True means visit all slots
    visits_are_sorted: bool = True              # True means visit basic types in alphabetial order, false in entry
    sort_class_slots: bool = False              # True means visit class slots in alphabetical order

    def __init__(self,
                 schema: Union[str, TextIO, SchemaDefinition, "Generator"],
                 format: Optional[str] = None,
                 emit_metadata: bool = False,
                 useuris: Optional[bool] = None,
                 importmap: Optional[str] = None,
                 log_level: int = DEFAULT_LOG_LEVEL_INT,
                 **kwargs) -> None:
        """
        Constructor

        :param schema: metamodel compliant schema.  Can be URI, file name, actual schema, another generator, an
        open file or a pre-parsed schema.
        :param format: expected output format
        :param emit_metadata: True means include date, generator, etc. information in source header if appropriate
        :param useuris: True means declared class slot uri's are used.  False means use model uris
        :param importmap: File name of import mapping file -- maps import name/uri to target
        :param log_level: Logging level
        """
        logging.basicConfig()
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(log_level)

        if format is None:
            format = self.valid_formats[0]
        assert format in self.valid_formats, f"Unrecognized format: {format}"
        self.format = format
        self.emit_metadata = emit_metadata
        if isinstance(schema, Generator):
            gen = schema
            self.schema = gen.schema
            self.synopsis = gen.synopsis
            self.loaded = gen.loaded
            self.namespaces = gen.namespaces
            self.base_dir = gen.base_dir
            self.importmap = gen.importmap
            self.schema_location = gen.schema_location
            self.schema_defaults = gen.schema_defaults
            self.logger = gen.logger
        else:
            loader = SchemaLoader(schema, self.base_dir, useuris=useuris, importmap=importmap, logger=self.logger)
            loader.resolve()
            self.schema = loader.schema
            self.synopsis = loader.synopsis
            self.loaded = loader.loaded
            self.namespaces = loader.namespaces
            self.base_dir = loader.base_dir
            self.importmap = loader.importmap
            self.schema_location = loader.schema_location
            self.schema_defaults = loader.schema_defaults

    def serialize(self, **kwargs) -> str:
        """
        Generate output in the required format

        :param kwargs: Generater specific parameters
        :return: Generated outputt
        """
        output = StringIO()
        with redirect_stdout(output):
            self.visit_schema(**kwargs)
            for sn, ss in (sorted(self.schema.subsets.items(), key=lambda s: s[0].lower()) if self.visits_are_sorted
                           else self.schema.subsets.items()):
                self.visit_subset(ss)
            for tn, typ in (sorted(self.schema.types.items(), key=lambda s: s[0].lower()) if self.visits_are_sorted
                            else self.schema.types.items()):
                self.visit_type(typ)
            for sn, slot in (sorted(self.schema.slots.items(), key=lambda c: c[0].lower()) if self.visits_are_sorted
                             else self.schema.slots.items()):
                self.visit_slot(self.aliased_slot_name(slot), slot)
            for cls in (sorted(self.schema.classes.values(), key=lambda c: c.name.lower()) if self.visits_are_sorted
                        else self.schema.classes.values()):
                if self.visit_class(cls):
                    for slot in self.all_slots(cls) if self.visit_all_class_slots else self.own_slots(cls):
                        self.visit_class_slot(cls, self.aliased_slot_name(slot), slot)
                    self.end_class(cls)
            self.end_schema(**kwargs)
        return output.getvalue()

    def visit_schema(self, **kwargs) -> None:
        """ Visited once at the beginning of generation

        @param kwargs: Arguments passed through from CLI -- implementation dependent
        """
        ...

    def end_schema(self, **kwargs) -> None:
        """ Visited once at the end of generation

        @param kwargs: Arguments passed through from CLI -- implementation dependent
        """
        ...

    def visit_class(self, cls: ClassDefinition) -> bool:
        """ Visited once per schema class

        @param cls: class being visited
        @return: Visit slots and end class.  False means skip and go on
        """
        return True

    def end_class(self, cls: ClassDefinition) -> None:
        """ Visited after visit_class_slots (if visit_class returned true)

        @param cls: class being visited
        """
        ...

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> None:
        """ Visited for each slot in a class.  If class level visit_all_slots is true, this is visited once
        for any class that is inherited (class itself, is_a, mixin, apply_to).  Otherwise just the own slots.

        @param cls: containing class
        @param aliased_slot_name: Aliased slot name.  May not be unique across all class slots
        @param slot: slot being visited
        """
        ...

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        """ Visited once for every slot definition in the schema.

        @param aliased_slot_name: Aliased name of the slot.  May not be unique
        @param slot: visited slot
        """
        ...

    def visit_type(self, typ: TypeDefinition) -> None:
        """ Visited once for every type definition in the schema

        @param typ: Type definition
        """
        ...

    def visit_subset(self, subset: SubsetDefinition) -> None:
        """ Visited once for every subset definition in the schema

        #param subset: Subset definition
        """
        ...

    # =============================
    # Helper methods
    # =============================
    def own_slots(self, cls: Union[ClassDefinitionName, ClassDefinition]) -> List[SlotDefinition]:
        """ Return the list of slots owned the class definition.  An "own slot" is any ``cls`` slot that does not appear
        in the class is_a parent.  Own_slots include:

            * any slot whose domain is cls
            * slot_usage entries
            * slots from mixins entries
            * slots from apply_to entries


        @param cls: class name or class definition name
        @return: list of owned slots.  List is sorted if sort_class_slots is true, otherwise in class order
        """
        if not isinstance(cls, ClassDefinition):
            cls = self.schema.classes[cls]
        parent = self.schema.classes[cls.is_a] if cls.is_a else None
        seen = set()
        rval = []
        for sname in cls.slots:
            sname_base = alias_root(self.schema, sname)
            if sname_base not in seen and (not parent or sname not in parent.slots):
                slot = self.schema.slots[sname]
                rval.append(slot)
                seen.add(sname_base)
        return sorted(rval, key=lambda s: s.name) if self.sort_class_slots else rval

    def own_slot_names(self, cls: Union[ClassDefinitionName, ClassDefinition]) -> List[SlotDefinitionName]:
        return [slot.name for slot in self.own_slots(cls)]

    def all_slots(self, cls: Union[ClassDefinitionName, ClassDefinition], *, cls_slots_first: bool = False,
                  seen: Optional[Set[ClassDefinitionName]] = None) \
            -> List[SlotDefinition]:
        """ Return all slots that are part of the class definition.  This includes all is_a, mixin and apply_to slots
        but does NOT include slot_usage targets.  If class B has a slot_usage entry for slot "s", only the slot
        definition for the redefined slot will be included, not its base.  Slots are added in the order they appear
        in classes, with recursive is_a's being added first followed by mixins and finally apply_tos

        @param cls: class definition or class definition name
        @param cls_slots_first: True means return own slots at the top of the list
        @param seen: List of slots already recorded. Used for internal recursion
        @return: ordered list of slots in the class with slot usages removed
        """
        if not isinstance(cls, ClassDefinition):
            cls = self.schema.classes[cls]
        if seen is None:
            seen = set()
        rval = []

        parent = self.schema.classes[cls.is_a] if cls.is_a else None
        if cls_slots_first:
            for slot in self.own_slots(cls):
                sname_base = alias_root(self.schema, slot.name)
                if sname_base not in seen:
                    rval.append(slot)
                    seen.add(sname_base)
            return rval + (self.all_slots(parent, cls_slots_first=cls_slots_first, seen=seen) if parent else [])
        else:
            for sname in cls.slots:
                sname_base = alias_root(self.schema, sname)
                if sname_base not in seen:
                    slot = self.schema.slots[sname]
                    rval.append(slot)
                    seen.add(sname_base)
            return sorted(rval, key=lambda s: s.name) if self.sort_class_slots else rval

    def parent(self, element: Union[ClassDefinition, SlotDefinition]) \
            -> Optional[Union[ClassDefinition, SlotDefinition]]:
        """ Return the parent of element, if any """
        return \
            None if element.is_a is None else \
            self.schema.classes[element.is_a] if isinstance(element, ClassDefinition) else \
            self.schema.slots[element.is_a]

    def ancestors(self, element: Union[ClassDefinition, SlotDefinition]) -> List[ElementName]:
        """ Return an ordered list of ancestor names for the supplied slot or class

        @param element: Slot or class name or definition
        @return: Ordered list of of ancestor names
        """
        return [element.name] + ([] if element.is_a is None else self.ancestors(self.parent(element)))

    def neighborhood(self, elements: Union[str, ElementName, List[ElementName]]) \
            -> References:
        """ Return a list of all slots, classes and types that touch any element in elements, including the element
        itself

        @param elements: Element names to do proximity with
        @return: All slots and classes that touch element
        """
        if isinstance(elements, (str, ElementName)):
            elements = [elements]
        touches = References()
        for element in elements:
            if element in self.schema.classes:
                touches.classrefs.add(cast(ClassDefinitionName, element))
                cls = self.schema.classes[cast(ClassDefinitionName, element)]
                if cls.is_a:
                    touches.classrefs.add(cls.is_a)
                # Mixins include apply_to's
                touches.classrefs.update(set(cls.mixins))
                for slotname in cls.slots:
                    slot = self.schema.slots[slotname]
                    if slot.range in self.schema.classes:
                        touches.classrefs.add(cast(ClassDefinitionName, slot.range))
                    elif slot.range in self.schema.types:
                        touches.typerefs.add(cast(TypeDefinitionName, slot.range))
                for cv in self.schema.classes.values():
                    if cv.is_a == element:
                        touches.classrefs.add(cv.name)
                if element in self.synopsis.rangerefs:
                    for slotname in self.synopsis.rangerefs[element]:
                        touches.slotrefs.add(slotname)
                        if self.schema.slots[slotname].domain:
                            touches.classrefs.add(self.schema.slots[slotname].domain)
                if cls.in_subset:
                    touches.subsetrefs.update(cls.in_subset)
            if element in self.schema.slots:
                touches.slotrefs.add(cast(SlotDefinitionName, element))
                slot = self.schema.slots[cast(SlotDefinitionName, element)]
                touches.slotrefs.update(set(slot.mixins))
                if slot.is_a:
                    touches.slotrefs.add(slot.is_a)
                if element in self.synopsis.inverses:
                    touches.slotrefs.update(self.synopsis.inverses[cast(SlotDefinitionName, element)])
                if slot.domain:
                    touches.classrefs.add(slot.domain)
                if slot.range in self.schema.classes:
                    touches.classrefs.add(cast(ClassDefinitionName, slot.range))
                elif slot.range in self.schema.types:
                    touches.typerefs.add(cast(TypeDefinitionName, slot.range))
                if slot.in_subset:
                    touches.subsetrefs.update(slot.in_subset)
                for sv in self.schema.slots.values():
                    if sv.is_a == element:
                        touches.slotrefs.add(sv.name)
            if element in self.schema.types:
                touches.typerefs.add(cast(TypeDefinitionName, element))
                typ = self.schema.types[cast(TypeDefinitionName, element)]
                if element in self.synopsis.rangerefs:
                    touches.slotrefs.update(self.synopsis.rangerefs[element])
                if typ.typeof:
                    touches.typerefs.add(cast(TypeDefinitionName, typ.typeof))
                if typ.in_subset:
                    touches.subsetrefs.update(typ.in_subset)
                for tv in self.schema.types.values():
                    if tv.typeof == element:
                        touches.slotrefs.add(tv.name)
            if element in self.schema.subsets:
                touches.subsetrefs.add(cast(SubsetDefinitionName, element))
                if element in self.synopsis.subsetrefs:
                    touches.update(self.synopsis.subsetrefs[cast(SubsetDefinitionName, element)])
            if not bool(touches):
                self.logger.warning(f"neighborhood({element}) - {element} is undefined")

        return touches

    def range_type_path(self, typ: TypeDefinition) -> List[str]:
        """
        Return a formatted list of range types from the base up

        :param typ: type definition whose name is to be formatted
        :return: List of possible types with base at the leftmost
        """
        formatted_typ_name = self.class_or_type_name(typ.name)
        if typ.typeof:
            return self.range_type_path(self.schema.types[cast(TypeDefinitionName, typ.typeof)]) + [formatted_typ_name]
        elif typ.repr:
            return [typ.repr, formatted_typ_name]
        else:
            return [formatted_typ_name]

    def class_identifier(self, def_or_name: Union[str, ClassDefinition, TypeDefinition], keys_count: bool = False) \
            -> Optional[SlotDefinitionName]:
        """
        Return the class identifier if any

        :param def_or_name: class name or definition
        :param keys_count: True means treat keys AND identifiers as identifiers
        :return: name of class identifier (or key) if one exists
        """
        if isinstance(def_or_name, ClassDefinition):
            cls = def_or_name
        elif def_or_name in self.schema.classes:
            cls = self.schema.classes[cast(ClassDefinitionName, def_or_name)]
        else:
            return None
        for slotname in cls.slots:
            slot = self.schema.slots[slotname]
            if slot.identifier or (keys_count and slot.key):
                return slotname
        return None

    def class_identifier_path(self, cls_or_clsname: Union[str, ClassDefinition], force_non_key: bool) -> List[str]:
        """
        Return the path closure to a class identifier if the class has a key and force_non_key is false otherwise
        return a dictionary closure.

        :param cls_or_clsname: class definition
        :param force_non_key: True means inlined even if the class has a key
        :return: path
        """
        cls = cls_or_clsname if isinstance(cls_or_clsname, ClassDefinition) \
            else self.schema.classes[cast(ClassDefinitionName, cls_or_clsname)]

        # Determine whether the class has a key
        identifier_slot = None
        if not force_non_key:
            identifier_slot = self.class_identifier(cls, keys_count=True)

        # No key or inlined, its closure is a dictionary
        if identifier_slot is None:
            return ['dict', self.class_or_type_name(cls.name)]

        # We're dealing with a reference
        pathname = camelcase(cls.name + ' ' + self.aliased_slot_name(identifier_slot))
        if cls.is_a:
            parent_identifier_slot = self.class_identifier(cls.is_a, keys_count=True)
            if parent_identifier_slot:
                return self.class_identifier_path(cls.is_a, False) + [pathname]
        return self.slot_type_path(identifier_slot) + [pathname]

    def slot_type_path(self, slot_or_name: Union[str, SlotDefinition]) -> List[str]:
        """
        Return a ordered list of types starting with the base type and ending with the most proximal type

        :param slot_or_name: slot whose range is being typed
        :return: ordered list of types from base type forward
        """
        slot = slot_or_name if isinstance(slot_or_name, SlotDefinition) \
            else self.schema.slots[cast(SlotDefinitionName, slot_or_name)]
        if slot.range in self.schema.types:
            # Type
            return self.range_type_path(self.schema.types[cast(TypeDefinitionName, slot.range)])
        else:
            # Class
            return self.class_identifier_path(slot.range, bool(slot.inlined))

    def aliased_slot_name(self, slot: Union[SlotDefinitionName, SlotDefinition]) -> SlotDefinitionName:
        """ Return the overloaded slot name -- the alias if one exists otherwise the actual name

        @param slot: either a slot name or a definition
        @return: overloaded name
        """
        if isinstance(slot, str):
            slot = self.schema.slots[cast(SlotDefinitionName, slot)]
        return slot.alias if slot.alias else slot.name

    def class_or_type_for(self, name: str) -> Optional[Element]:
        """
        Return the corresponding class or type for name
        """
        if name in self.schema.classes:
            return self.schema.classes[name]
        elif name in self.schema.types:
            return self.schema.types[cast(TypeDefinitionName, name)]
        return None

    def class_or_type_name(self, name: str) -> str:
        """
        Return the camelcase representation of clsname if it is a valid class or type.  Prepend "Unknown"
        if the name isn't valid
        """
        if name in self.schema.classes:
            return camelcase(name)
        elif name in self.schema.types:
            typ = self.schema.types[cast(TypeDefinitionName, name)]
            if typ.typeof:
                return camelcase(name)
            else:
                return typ.base
        else:
            return "Unknown_" + camelcase(name)

    def slot_for(self, name: str) -> Optional[Element]:
        return self.schema.slots.get(name)

    def slot_name(self, name: str) -> str:
        """
        Return the underscored version of the aliased slot name if name is a slot. Prepend "unknown_" if the name
        isn't valid.
        """
        slot = self.slot_for(name)
        return underscore(self.aliased_slot_name(slot) if slot else ("unknown " + name))

    def subset_for(self, name: str) -> Optional[Element]:
        return self.schema.subsets.get(name)

    def subset_name(self, name: str) -> str:
        subset = self.subset_for(name)
        return ('' if subset else "Unknown_") + camelcase(name)

    def formatted_element_name(self, el_or_elname: Union[ElementName, Element],
                               is_range_name: bool = False) -> Optional[str]:
        """
        Return the default format for the name or the referenced element.  Slots are under_scored, all others are
        CamelCased. Slot names are the alias, not the actual name

        :param el_or_elname: element or name to map
        :param is_range_name: True means that we're looking for a class or type.  False means Slot or Subset. Only
        applies if el_or_elname is an ElementName (otherwise we know what we've got
        :return: Formatted name if type can be known else None
        """
        if isinstance(el_or_elname, str):
            if is_range_name:
                return self.class_or_type_name(el_or_elname) \
                    if el_or_elname in self.schema.classes or \
                       el_or_elname in self.schema.types or \
                       el_or_elname == self.schema.default_range else None
            elif el_or_elname in self.schema.slots:
                return self.slot_name(cast(SlotDefinitionName, el_or_elname))
            elif el_or_elname in self.schema.subsets:
                return self.subset_name(el_or_elname)
            return None
        elif isinstance(el_or_elname, (ClassDefinition, TypeDefinition)):
            return self.class_or_type_name(el_or_elname.name)
        elif isinstance(el_or_elname, SlotDefinition):
            return self.slot_name(el_or_elname.name)
        elif isinstance(el_or_elname, SubsetDefinition):
            return self.subset_name(el_or_elname.name)
        else:
            return None

    def obj_for(self, el_or_elname: str, is_range_name: bool = False) -> Optional[Element]:
        if is_range_name:
            return self.class_or_type_for(el_or_elname) \
                if el_or_elname in self.schema.classes or \
                   el_or_elname in self.schema.types or \
                   el_or_elname == self.schema.default_range else None
        elif el_or_elname in self.schema.slots:
            return self.slot_for(cast(SlotDefinitionName, el_or_elname))
        elif el_or_elname in self.schema.subsets:
            return self.subset_for(el_or_elname)
        else:
            return self.class_or_type_for(el_or_elname)

    def default_prefix(self) -> Optional[str]:
        """ Return the default prefix for the schema

         @return: URI or NCNAME of default prefix """
        if '://' in self.schema.default_prefix:
            return self.schema.default_prefix
        else:
            # Basic loader tests for valid default prefix
            return self.schema.prefixes[PrefixPrefixPrefix(self.schema.default_prefix)].prefix_reference

    # TODO: add lru cache once we get identity into the classes
    def domain_slots(self, cls: ClassDefinition) -> List[SlotDefinition]:
        """ Return all slots in the class definition that are owned by the class """
        return [slot for slot in [self.schema.slots[sn] for sn in cls.slots] if slot.owner == cls.name]


def shared_arguments(g: Generator) -> Callable[[Command], Command]:
    _LOG_LEVEL_STRINGS = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']

    def _log_level_string_to_int(log_level_string: str) -> int:
        log_level_string = log_level_string.upper()
        level = [e for e in log_level_string if e.startswith(log_level_string)]
        if not level:
            pass
        log_level_int = getattr(logging, log_level_string[0], logging.INFO)
        assert isinstance(log_level_int, int)
        return log_level_int

    def decorator(f: Command) -> Command:
        f.params.append(
            Argument(("yamlfile", ), type=click.Path(exists=True, dir_okay=False)))
        f.params.append(
            Option(("--format", "-f"), type=click.Choice(g.valid_formats), help="Output format",
                   default=g.valid_formats[0]))
        f.params.append(
            Option(("--metadata/--no-metadata", ), default=True, help="Include metadata in output"))
        f.params.append(
            Option(("--useuris/--metauris", ), default=True, help="Include metadata in output"))
        f.params.append(
            Option(("--importmap", "-im"), type=click.File(), help="Import mapping file")
        )
        f.params.append(
            Option(("--log_level", ), type=click.Choice(_LOG_LEVEL_STRINGS),
                   help=f"Logging level (default={DEFAULT_LOG_LEVEL})",
                   default=DEFAULT_LOG_LEVEL)
        )
        return f
    return decorator





import os
import re
import unittest

from functools import reduce
from typing import List, Tuple

from rdflib import Graph

from biolinkml.generators.owlgen import OwlSchemaGenerator
from tests.test_utils import inputdir
from tests.test_utils.support.base import Base

repl: List[Tuple[str, str]] = [
    (r'\s*meta:generation_date ".*" ;', 'meta:generation_date "Fri Jan 25 14:22:29 2019" ;'),
    (r'\s*meta:source_file_date ".*" ;', 'meta:source_file_date "Fri Jan 25 14:22:29 2019" ;')
]


def filtr(txt: str) -> str:
    return reduce(lambda s, expr: re.sub(expr[0], expr[1], s, flags=re.MULTILINE), repl, txt)


class OWLTestCase(Base):

    def assertOwlEqual(self, s1: str, s2: str) -> None:
        g1 = Graph()
        g2 = Graph()
        g1.parse(data=s1, format="turtle")
        g2.parse(data=s2, format="turtle")
        self.assertTrue(g1.isomorphic(g2))

    def test_cardinalities(self):
        owl_txt = OwlSchemaGenerator(os.path.join(inputdir, 'owl1.yaml')).serialize()
        self.eval_output(owl_txt, 'owl1.owl', filtr, self.assertOwlEqual)

    def test_pred_types(self):
        owl_txt = OwlSchemaGenerator(os.path.join(inputdir, 'owl2.yaml')).serialize()
        self.eval_output(owl_txt, 'owl2.owl', filtr, self.assertOwlEqual)


if __name__ == '__main__':
    unittest.main()



import datetime
import os
import unittest
from json import loads
from typing import Optional, Callable, Any

from jsonasobj import as_json

from biolinkml.meta import SchemaDefinition
from biolinkml.utils.schemaloader import SchemaLoader
from tests import sourcedir
from tests.test_utils import outputdir, inputdir

update_all_files: bool = False


class Base(unittest.TestCase):
    def fix_schema_metadata(self, schema: SchemaDefinition) -> SchemaDefinition:
        self.assertIsNotNone(schema.generation_date)
        schema.source_file = os.path.basename(schema.source_file)
        schema.generation_date = "2018-12-31 17:23"
        self.assertIsNotNone(schema.metamodel_version)
        schema.metamodel_version = "0.5.0"
        self.assertIsNotNone(schema.source_file_size)
        schema.source_file_size = 259
        self.assertIsNotNone(schema.source_file_date)
        schema.source_file_date = "2018-12-31 17:23"
        return schema


    def eval_output(self, actual: str, filename: str, conv_f: Optional[Callable[[str], Any]] = None,
                    comp_f: Optional[Callable[[str, str], None]] = None) -> None:

        def _default_comparator(expected: str, actual: str) -> None:
            if expected != actual:
                print(f"\n***** Testing file: {file_path}. Remove to update test. *****\n")
                self.assertEqual(expected, actual)

        if comp_f is None:
            comp_f = _default_comparator

        file_path = os.path.join(outputdir, filename)
        file_created = False
        if not os.path.exists(file_path) or update_all_files:
            print(f"Creating {file_path}")
            with open(file_path, 'w') as f:
                f.write(actual)
            file_created = True
        with open(file_path) as f:
            expected = f.read()

        if conv_f:
            self.maxDiff = None
            comp_f(conv_f(expected), conv_f(actual))
        else:
            comp_f(expected, actual)
        self.assertFalse(file_created, f"{file_path}: Created new file image -- rerun")

    def eval_loader(self, base_name: str, is_sourcedir: bool=False, source: Optional[str]=None) -> None:
        fn = os.path.join(sourcedir if is_sourcedir else inputdir, base_name + '.yaml') if not source else source
        loader = SchemaLoader(fn)
        schema = as_json(self.fix_schema_metadata(loader.resolve()))
        self.eval_output(schema, base_name + '.json', loads)
        errors = '\n'.join(loader.synopsis.errors())
        self.eval_output(errors, base_name + '.errs')
        self.assertFalse(update_all_files, "Updating base files -- rerun")



import datetime
import unittest
from dataclasses import dataclass

from jsonasobj import as_json
from rdflib import Literal, XSD, Graph, RDF, Namespace
from rdflib.term import Identifier

from biolinkml.utils.metamodelcore import NCName, Bool, URIorCURIE, URI, XSDDate, XSDDateTime, XSDTime, Curie, \
    NodeIdentifier
from biolinkml.utils.namespaces import Namespaces
from biolinkml.utils.strictness import lax, strict
from biolinkml.utils.yamlutils import YAMLRoot, as_rdf


class MetamodelCoreTest(unittest.TestCase):
    def tearDown(self) -> None:
        strict()

    def test_ncname(self):
        self.assertEqual('A123', NCName('A123'))
        x = NCName('A1.B_C-')
        self.assertEqual('A1.B_C-', x)
        self.assertIsInstance(x, str)
        self.assertIsInstance(x, NCName)
        self.assertEqual(x, NCName(x))
        x = str(x)
        self.assertIsInstance(x, str)
        self.assertFalse(isinstance(x, NCName))
        with self.assertRaises(ValueError):
            NCName('1')
        with self.assertRaises(ValueError):
            NCName('A12!')

    def test_uris(self):
        """ Test the URI and URIorCURIE types """
        str1 = "https://google.com/test#file?abc=1&def=4"
        self.assertEqual(str1, URIorCURIE(str1))
        self.assertEqual(str1, URI(str1))
        str2 = "abc:123"
        self.assertEqual(str2, URIorCURIE(str2))
        str3 = ":123"
        self.assertEqual(str3, URIorCURIE(str3))
        with self.assertRaises(ValueError):
            URI(str2)
        with self.assertRaises(ValueError):
            URIorCURIE("1abc:def")
        with self.assertRaises(ValueError):
            URIorCURIE("1:def")
        lax()
        URI(str2)
        URIorCURIE("1abc:def")
        URIorCURIE("1:def")

    def test_curie(self):
        """ Test the CURIE type """
        self.assertEqual("rdf:type", Curie("rdf:type"))
        with self.assertRaises(ValueError):
            Curie("type")
        self.assertFalse(Curie.is_valid("type"))
        self.assertEqual(":type", Curie(":type"))
        self.assertTrue(Curie.is_valid(':type'))
        with self.assertRaises(ValueError):
            Curie("1df:type")
        self.assertFalse(Curie.is_valid('1df:type'))
        with self.assertRaises(ValueError):
            Curie("rdf:17")
        self.assertFalse(Curie.is_valid('rdf:17'))
        nsm = Namespaces(Graph())
        self.assertEqual(RDF.type, Curie("rdf:type").as_uri(nsm))
        self.assertIsNone(Curie("ex:foo").as_uri(nsm))
        self.assertIsNone(Curie(":bar").as_uri(nsm))
        nsm._default = "http://example.org/test#"
        self.assertEqual(nsm._default['bear'], Curie(":bear").as_uri(nsm))


    def test_uri(self):
        """ Test the URI data type """
        self.assertEqual("http://foo.org/bargles", URI("http://foo.org/bargles"))
        with self.assertRaises(ValueError):
            URI("rdf:type")
        lax()
        x = URI("rdf:type")
        self.assertFalse(URI.is_valid(x))

    def test_bool(self):
        self.assertTrue(Bool(True))
        self.assertTrue(Bool("True"))
        self.assertTrue(Bool("true"))
        self.assertTrue(Bool(1))
        self.assertTrue(Bool("1"))
        self.assertTrue(Bool(Bool(True)))
        self.assertFalse(Bool(False))
        self.assertFalse(Bool("False"))
        self.assertFalse(Bool("false"))
        self.assertFalse(Bool(0))
        self.assertFalse(Bool("0"))
        self.assertFalse(Bool(Bool(0)))
        # Strict mode
        with self.assertRaises(ValueError):
            x = Bool(17)
        with self.assertRaises(ValueError):
            x = Bool("a")
        lax()
        x = Bool(17)
        self.assertFalse(Bool.is_valid(17))
        x = Bool("a")
        self.assertFalse(Bool.is_valid("a"))
        self.assertTrue(Bool.is_valid(True))
        self.assertTrue(Bool.is_valid(Bool(True)))

    def test_time(self):
        v = datetime.time(13, 17, 43, 279000)
        self.assertEqual('13:17:43.279000', XSDTime(v))                              # A date can be a datetime
        self.assertEqual('13:17:43.279000', XSDTime(Literal(v, datatype=XSD.time)))  # An RDFLIB Literal
        self.assertEqual('13:17:43.279000', v.isoformat())                           # A string
        self.assertEqual('13:17:43.279000', XSDTime(XSDTime(v)))                     # An existing date
        strict()
        with self.assertRaises(ValueError):
            XSDTime('Jan 12, 2019')
        with self.assertRaises(ValueError):
            XSDTime(datetime.datetime.now())
        lax()
        self.assertEqual('Jan 12, 2019', XSDTime('Jan 12, 2019'))
        XSDDate(datetime.datetime.now())
        self.assertFalse(XSDTime.is_valid('Jan 12, 2019'))
        self.assertFalse(XSDTime.is_valid(datetime.datetime.now()))
        self.assertTrue(XSDTime.is_valid(v))

    def test_date(self):
        v = datetime.date(2019, 7, 6)
        self.assertEqual('2019-07-06', XSDDate(v))                              # A date can be a datetime
        self.assertEqual('2019-07-06', XSDDate(Literal(v, datatype=XSD.date)))  # An RDFLIB Literal
        self.assertEqual('2019-07-06', v.isoformat())                           # A string
        self.assertEqual('2019-07-06', XSDDate(XSDDate(v)))                     # An existing date
        strict()
        with self.assertRaises(ValueError):
            XSDDate('Jan 12, 2019')
        with self.assertRaises(ValueError):
            XSDDate(datetime.datetime.now())
        lax()
        bv = XSDDate('Jan 12, 2019')
        self.assertEqual('Jan 12, 2019', bv)
        self.assertFalse(XSDDate.is_valid(bv))
        XSDDate(datetime.datetime.now())
        self.assertFalse(XSDDate.is_valid('Jan 12, 2019'))
        self.assertFalse(XSDDate.is_valid(datetime.datetime.now()))
        self.assertTrue(XSDDate.is_valid(XSDDate(datetime.datetime.now().date())))
        self.assertTrue(XSDDate.is_valid(v))

    def test_datetime(self):
        v = datetime.datetime(2019, 7, 6, 17, 22, 39, 7300)
        self.assertEqual('2019-07-06T17:22:39.007300', XSDDateTime(v))
        self.assertEqual('2019-07-06T17:22:39.007300', XSDDateTime(Literal(v, datatype=XSD.datetime)))
        self.assertEqual('2019-07-06T17:22:39.007300', XSDDateTime(Literal(v).value))
        self.assertEqual('2019-07-06T17:22:39.007300', v.isoformat())
        self.assertEqual('2019-07-06T17:22:39.007300', XSDDateTime(XSDDateTime(v)))
        vstr = str(Literal(v).value)
        self.assertEqual('2019-07-06 17:22:39.007300', vstr)       # Note that this has no 'T'
        self.assertEqual('2019-07-06T17:22:39.007300', XSDDateTime(vstr))
        with self.assertRaises(ValueError):
            XSDDateTime('Jan 12, 2019')
        lax()
        self.assertEqual('penguins', XSDDateTime('penguins'))
        XSDDateTime(datetime.datetime.now())
        self.assertFalse(XSDDateTime.is_valid('Jan 12, 2019'))
        self.assertTrue(XSDDateTime.is_valid(datetime.datetime.now()))
        self.assertTrue(XSDDateTime.is_valid(XSDDate(datetime.datetime.now().date())))
        self.assertTrue(XSDDateTime.is_valid(v))

    @unittest.skipIf(True, "Finish implementing this")
    def test_nodeidentifier(self):
        context = """{
            "@context": {
                "type": "@type",
                "OIO": "http://www.geneontology.org/formats/oboInOwl#",
                "dcterms": "http://purl.org/dc/terms/",
                "metatype": "https://w3id.org/biolink/biolinkml/type/",
                "owl": "http://www.w3.org/2002/07/owl#",
                "pav": "http://purl.org/pav/",
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "skos": "http://www.w3.org/2004/02/skos/core#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "s": {
                    "@id": "rdfs:subject"
                },
                "t": {
                    "@id": "rdfs:object"
                }
            }
        }"""
        EX = Namespace("http://example.org/tests/")
        class Root(NodeIdentifier):
            pass

        class Child1(Root):
            pass

        class Child2(Root):
            pass

        class Desc1(Child1):
            pass

        class Unassoc(NodeIdentifier):
            pass

        @dataclass
        class Pair(YAMLRoot):
            s: Child1 = None
            t: Child2 = None

            def __post_init__(self):
                if not isinstance(self.s, Child1):
                    self.s = Child1(self.s)
                if not isinstance(self.t, Child2):
                    self.t = Child2(self.t)
                    super().__post_init__()

        s = Desc1(EX.descendant1)
        t = Child2(EX.child2)
        y = Pair(s, t)
        self.assertEqual("""{
   "s": "http://example.org/tests/descendant1",
   "t": "http://example.org/tests/child2"
}""", as_json(y))
        self.assertEqual("""@prefix OIO: <http://www.geneontology.org/formats/oboInOwl#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix metatype: <https://w3id.org/biolink/biolinkml/type/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix pav: <http://purl.org/pav/> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a <file:///Users/solbrig/git/biolink/biolinkml/tests/Pair> ;
    rdfs:object "http://example.org/tests/child2" ;
    rdfs:subject "http://example.org/tests/descendant1" .

""", as_rdf(y, context).serialize(format="turtle").decode())
        with self.assertRaises(ValueError):
            y = Pair(s, s)




if __name__ == '__main__':
    unittest.main()


from typing import Optional

from rdflib import Namespace


class CurieNamespace(Namespace):
    def __new__(cls, prefix: str, value):
        try:
            rt = str.__new__(cls, value)
        except UnicodeDecodeError:
            rt = str.__new__(cls, value, 'utf-8')
        rt.prefix = prefix
        return rt

    def curie(self, reference: Optional[str] = '') -> str:
        return self.prefix + ':' + reference



import os
import unittest

from biolinkml.generators.shexgen import ShExGenerator
from tests.test_utils import inputdir


class URLImportTestCase(unittest.TestCase):

    @unittest.skipIf(False, "Finish implementing this")
    def test_import_from_url(self):
        """ Validate namespace bindings """
        shex = ShExGenerator(os.path.join(inputdir, 'import_test_l2.yaml')).serialize()
        self.assertEqual("""BASE <http://example.org/l2/>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX l1: <http://example.org/l1/>
PREFIX base: <http://example.org/b/>


l1:Int xsd:integer

base:String xsd:string

base:BaseClass CLOSED {
    (  $base:BaseClass_tes base:base_slot @base:String ? ;
       rdf:type [ base:BaseClass ] ?
    )
}

l1:L1Class  (
    CLOSED {
       (  $l1:L1Class_tes (  l1:l1_slot1 @base:String ? ;
             l1:l1_slot2 @l1:Int ?
          ) ;
          rdf:type [ l1:L1Class ] ?
       )
    } OR @<L2Class>
)

<L2Class> CLOSED {
    (  $<L2Class_tes> (  &l1:L1Class_tes ;
          rdf:type [ l1:L1Class ] ? ;
          <l2_slot1> @base:String ? ;
          <l2_slot2> @l1:Int ?
       ) ;
       rdf:type [ <L2Class> ] ?
    )
}""", shex.strip())


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from biolinkml.generators.yumlgen import YumlGenerator
from tests.test_issues import sourcedir


class Issue12UnitTest(unittest.TestCase):
    def test_domain_slots(self):
        """ has_phenotype shouldn't appear in the UML graph """
        yuml = YumlGenerator(os.path.join(sourcedir, 'issue_12.yaml')).serialize()
        self.assertEqual('http://yuml.me/diagram/nofunky;dir:TB/class/', yuml)


if __name__ == '__main__':
    unittest.main()



"""
Generate JSON-LD contexts

"""
import logging
import os
from typing import Union, TextIO, Set, Optional

import click
from jsonasobj import JsonObj, as_json
from rdflib import XSD

from biolinkml.meta import SchemaDefinition, ClassDefinition, SlotDefinition, Definition, Element
from biolinkml.utils.formatutils import camelcase, underscore, be
from biolinkml.utils.generator import Generator, shared_arguments
from includes.types import SHEX

URI_RANGES = (XSD.anyURI, SHEX.nonliteral, SHEX.bnode, SHEX.iri)


class ContextGenerator(Generator):
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ['json']
    visit_all_class_slots = False

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], **kwargs) -> None:
        super().__init__(schema, **kwargs)
        if self.namespaces is None:
            raise TypeError("Schema text must be supplied to context generater.  Preparsed schema will not work")
        self.emit_prefixes: Set[str] = set()
        self.default_ns = None
        self.context_body = dict()
        self.slot_class_maps = dict()

    def visit_schema(self, base: Optional[str]=None, output: Optional[str]=None, **_):
        # Add any explicitly declared prefixes
        for prefix in self.schema.prefixes.values():
            self.emit_prefixes.add(prefix.prefix_prefix)

        # Add any prefixes explicitly declared
        for pfx in self.schema.emit_prefixes:
            self.add_prefix(pfx)

        # Add the default prefix
        if self.schema.default_prefix:
            dflt = self.namespaces.prefix_for(self.schema.default_prefix)
            if dflt:
                self.default_ns = dflt
            default_uri = self.namespaces[self.default_ns]
            self.emit_prefixes.add(self.default_ns)
            self.context_body['@vocab'] = default_uri
            # self.context_body['@base'] = self.base_dir

    def end_schema(self, base: Optional[str] = None, output: Optional[str] = None, **_) -> None:
        comments = f'''Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}
Generation date: {self.schema.generation_date}
Schema: {self.schema.name}

id: {self.schema.id}
description: {be(self.schema.description)}
license: {be(self.schema.license)}
'''
        context = JsonObj()
        context["_comments"] = comments
        context_content = {"type": "@type"}
        if base:
            if '://' not in base:
                self.context_body['@base'] = os.path.relpath(base, os.path.dirname(self.schema.source_file))
            else:
                self.context_body['@base'] = base
        for prefix in sorted(self.emit_prefixes):
            context_content[prefix] = self.namespaces[prefix]
        for k, v in self.context_body.items():
            context_content[k] = v
        for k, v in self.slot_class_maps.items():
            context_content[k] = v
        context['@context'] = context_content
        if output:
            with open(output, 'w') as outf:
                outf.write(as_json(context))
        else:
            print(as_json(context))

    def visit_class(self, cls: ClassDefinition) -> bool:
        class_def = {}
        cn = camelcase(cls.name)
        self.add_mappings(cls)
        cls_prefix = self.namespaces.prefix_for(cls.class_uri)
        if not self.default_ns or not cls_prefix or cls_prefix != self.default_ns:
            class_def['@id'] = cls.class_uri
            if cls_prefix:
                self.add_prefix(cls_prefix)
        if class_def:
            self.slot_class_maps[cn] = class_def

        # We don't bother to visit class slots - just all slots
        return False

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        if slot.identifier:
            slot_def = '@id'
        else:
            slot_def = {}
            if not slot.is_usage_slot:
                if slot.range in self.schema.classes:
                    slot_def['@type'] = '@id'
                else:
                    range_type = self.schema.types[slot.range]
                    if self.namespaces.uri_for(range_type.uri) == XSD.string:
                        pass
                    elif self.namespaces.uri_for(range_type.uri) in URI_RANGES:
                        slot_def['@type'] = '@id'
                    else:
                        slot_def['@type'] = range_type.uri
                slot_prefix = self.namespaces.prefix_for(slot.slot_uri)
                if not self.default_ns or not slot_prefix or slot_prefix != self.default_ns:
                    slot_def['@id'] = slot.slot_uri
                    if slot_prefix:
                        self.add_prefix(slot_prefix)
                self.add_mappings(slot)
        if slot_def:
            self.context_body[underscore(aliased_slot_name)] = slot_def

    def add_prefix(self, ncname: str) -> None:
        """ Add a prefix to the list of prefixes to emit

        @param ncname: name to add
        """
        if ncname not in self.namespaces:
            self.logger.warning(f"Unrecognized prefix: {ncname}")
            self.namespaces[ncname] = f"http://example.org/UNKNOWN/{ncname}/"
        self.emit_prefixes.add(ncname)

    def add_mappings(self, defn: Definition) -> None:
        """
        Process any mappings in defn, adding all of the mappings prefixes to the namespace map
        :param defn: Class or Slot Definition
        """
        self.add_id_prefixes(defn)
        for mapping in defn.mappings:
            if '://' in mapping:
                mcurie = self.namespaces.curie_for(mapping)
                self.logger.warning(f"No namespace defined for URI: {mapping}")
                if mcurie is None:
                    return        # Absolute path - no prefix/name
                else:
                    mapping = mcurie
            if ':' not in mapping or len(mapping.split(':')) != 2:
                raise ValueError(f"Definition {defn.name} - unrecognized mapping: {mapping}")
            ns = mapping.split(':')[0]
            self.add_prefix(ns)

    def add_id_prefixes(self, element: Element) -> None:
        for id_prefix in element.id_prefixes:
            self.add_prefix(id_prefix)


@shared_arguments(ContextGenerator)
@click.command()
@click.option("-o", "--output", help="Output file name")
@click.option("--base", help="Base URI for model")
@click.option("--native_uris", is_flag=True, help="Don't use the class and slot URI's")
def cli(yamlfile, **args):
    """ Generate jsonld @context definition from biolink model """
    print(ContextGenerator(yamlfile, **args).serialize(**args))



import unittest

from biolinkml.generators.csvgen import cli
from tests import source_yaml_path
from tests.test_scripts.clicktestcase import ClickTestCase

# This has to occur post ClickTestCase
import click


class GenCSVTestCase(ClickTestCase):
    testdir = "gencsv"
    click_ep = cli
    prog_name = "gen-csv"

    def test_help(self):
        self.do_test("--help", 'help')

    def test_meta(self):
        self.do_test(source_yaml_path, 'meta.csv')
        self.do_test(source_yaml_path + ' -f tsv', 'meta.tsv')
        self.do_test(source_yaml_path + ' -f xsv', 'meta_error', error=click.exceptions.BadParameter)
        self.do_test([source_yaml_path, "-r", "schema_definition"], 'meta_sd')
        self.do_test([source_yaml_path, "-r", "schema_definition", "-r", "slot_definition"], 'meta_sd_sd')
        self.do_test([source_yaml_path, "-r", "nada"], 'meta_sd', error=ValueError)


if __name__ == '__main__':
    unittest.main()



import os
import sys
import unittest
from dataclasses import dataclass
from typing import List, Optional, Union, Tuple

import requests

from rdflib import Namespace, URIRef

W3ID_SERVER = "https://w3id.org/"
DEFAULT_SERVER = W3ID_SERVER
# DEFAULT_SERVER = "http://localhost:8091/"

# Taken from Firefox network.http.accept.default
default_header = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
github_io = "https://biolink.github.io/"


@dataclass
class TestEntry:
    input_url: Union[str, URIRef, Namespace]
    expected_url: str
    accept_header: Optional[str] = None


class TestLists:
    def __init__(self, server: str) -> None:
        if not server.endswith(('#', '/')):
            server += '/'
        self.biolink = server + 'biolink/'
        self.biolinkml = self.biolink + 'biolinkml/'
        self.types = Namespace(self.biolinkml + 'types')
        self.mappings = Namespace(self.biolinkml + 'mappings')
        self.metas = Namespace(self.biolinkml + 'meta')
        self.type = Namespace(self.biolinkml + 'type/')
        self.mapping = Namespace(self.biolinkml + 'mapping/')
        self.meta = Namespace(self.biolinkml + 'meta/')

        self.biolink_general: List[TestEntry] = [
            TestEntry(self.biolink, 'biolink-model/'),
            TestEntry(self.biolink + 'foo', 'biolink-model/foo')
        ]

        self.model_entries: List[TestEntry] = [
            TestEntry(self.types, 'biolinkml/includes/types'),
            TestEntry(self.types, 'biolinkml/includes/types.yaml', 'text/yaml'),
            TestEntry(self.types, 'biolinkml/includes/types.ttl', 'text/turtle'),
            TestEntry(self.types, 'biolinkml/includes/types.jsonld', 'application/json'),
            TestEntry(self.types, 'biolinkml/includes/types.shex', 'text/shex'),
            TestEntry(self.types['.owl'], 'biolinkml/includes/types.owl'),
            TestEntry(self.biolinkml + 'includes/context.jsonld', 'biolinkml/includes/context.jsonld'),
            TestEntry(self.types['/'], 'biolinkml/includes/types/'),
            TestEntry(self.mappings, 'biolinkml/includes/mappings'),
            TestEntry(self.mappings, 'biolinkml/includes/mappings.yaml', 'text/yaml'),
            TestEntry(self.mappings, 'biolinkml/includes/mappings.ttl', 'text/turtle'),
            TestEntry(self.mappings, 'biolinkml/includes/mappings.jsonld', 'application/json'),
            TestEntry(self.mappings, 'biolinkml/includes/mappings.shex', 'text/shex'),
            TestEntry(self.mappings['.owl'], 'biolinkml/includes/mappings.owl'),
            # TestEntry(self.mappings, 'biolinkml/includes/mappings.context.jsonld', 'application.jsonld'),
            TestEntry(self.mappings['/'], 'biolinkml/includes/mappings/')
        ]

        self.vocab_entries: List[TestEntry] = [
            TestEntry(self.type['index'], 'biolinkml/docs/types/index'),
            TestEntry(self.type.Element, 'biolinkml/docs/types/Element.yaml', 'text/yaml'),
            TestEntry(self.type.Element, 'biolinkml/docs/types/Element.ttl', 'text/turtle'),
            TestEntry(self.type.slots, 'biolinkml/docs/types/slots.jsonld', 'application/json'),
            TestEntry(self.mapping['index'], 'biolinkml/docs/mappings/index')
        ]

        self.meta_model_entries: List[TestEntry] = [
            TestEntry(self.metas, 'biolinkml/meta'),
            TestEntry(self.metas, 'biolinkml/meta.yaml', 'text/yaml'),
            TestEntry(self.metas, 'biolinkml/meta.ttl', 'text/turtle'),
            TestEntry(self.metas, 'biolinkml/meta.jsonld', 'application/json'),
            TestEntry(self.metas, 'biolinkml/meta.shex', 'text/shex'),
            TestEntry(self.metas['.owl'], 'biolinkml/meta.owl'),
            TestEntry(self.metas['.foo'], 'biolinkml/meta.foo'),
            TestEntry(self.biolinkml + 'context.jsonld', 'biolinkml/context.jsonld'),
            TestEntry(self.biolinkml + 'contextn.jsonld', 'biolinkml/contextn.jsonld')
        ]
        self.meta_vocab_entries: List[TestEntry] = [
            TestEntry(self.meta.Element, 'biolinkml/docs/Element'),
            TestEntry(self.meta.slot, 'biolinkml/docs/slot'),
            TestEntry(self.meta.Element, 'biolinkml/docs/Element.yaml', 'text/yaml'),
            TestEntry(self.meta.Element, 'biolinkml/docs/Element.ttl', 'text/turtle'),
            TestEntry(self.meta.Element, 'biolinkml/docs/Element.jsonld', 'application/json'),
            TestEntry(self.meta.Element, 'biolinkml/docs/Element.shex', 'text/shex'),
            TestEntry(self.meta.Element, 'biolinkml/docs/Element', 'text/foo'),
        ]

        self.biolinkmodel_entries: List[TestEntry] = [
            TestEntry(self.biolink + 'vocab/Element', 'biolink-model/docs/Element'),
            TestEntry(self.biolink + 'vocab', 'biolink-model/docs'),
            TestEntry(self.biolink + 'context.jsonld', 'biolink-model/context.jsonld'),
            TestEntry(self.biolink + 'contextn.jsonld', 'biolink-model/contextn.jsonld'),
            TestEntry(self.biolink + 'biolink-model.yaml', 'biolink-model/biolink-model.yaml')
        ]


FAIL_ON_ERROR = True


@unittest.skipIf(True, "Rewrite rules test bypassed except when new rules are submitted to w3id.org")
class RewriteRuleTestCase(unittest.TestCase):
    SERVER = DEFAULT_SERVER         # Can be overwritten with a startup parameter

    @classmethod
    def setUpClass(cls):
        cls.tests = TestLists(cls.SERVER)
        print(f"Server: {cls.SERVER}")
        cls.results: Tuple[str, str, str] = set()   # from, to, format

    @classmethod
    def tearDownClass(cls):
        print()
        for from_url, to_url, hdr in sorted(list(cls.results)):
            fmt = '' if hdr == 'text/html' else f" ({hdr})"
            if DEFAULT_SERVER != W3ID_SERVER:
                from_url = from_url.replace(DEFAULT_SERVER, W3ID_SERVER)
            print(f"{from_url}{fmt} - {to_url}")

    def record_results(self, from_url: str, accept_header, to_url: str) -> None:
        self.results.add( (from_url, to_url, accept_header.split(',')[0]))

    def rule_test(self, entries: List[TestEntry]) -> None:

        def test_it(e: TestEntry, accept_header: str) -> bool:
            expected = github_io + e.expected_url
            resp = requests.head(e.input_url, headers={'accept': accept_header}, verify=False)

            # w3id.org uses a 301 to go from http: to https:
            if resp.status_code == 301 and 'location' in resp.headers:
                resp = requests.head(resp.headers['location'], headers={'accept': accept_header}, verify=False)
            actual = resp.headers['location'] \
                if resp.status_code == 302 and 'location' in resp.headers \
                else f"Error: {resp.status_code}"
            if FAIL_ON_ERROR:
                self.assertEqual(expected, actual, f"redirect for: {resp.url}")
                self.record_results(e.input_url, accept_header, actual)
                return True
            elif expected != actual:
                print(f"{e.input_url} ({accept_header}):\n expected {expected} - got {actual}")
                return False
            self.record_results(e.input_url, accept_header, actual)
            return True

        def ev_al(entry: TestEntry) -> bool:
            if not entry.accept_header:
                return test_it(entry, default_header)
            else:
                r1 = test_it(entry, entry.accept_header)
                return test_it(entry, entry.accept_header + ',' + default_header) and r1

        self.assertTrue(all([ev_al(entry) for entry in entries]))

    def test_type_model(self):
        self.rule_test(self.tests.model_entries)

    def test_type_entry(self):
        self.rule_test(self.tests.vocab_entries)

    def test_meta_model(self):
        self.rule_test(self.tests.meta_model_entries)

    def test_meta_entry(self):
        self.rule_test(self.tests.meta_vocab_entries)

    def test_general(self):
        self.rule_test(self.tests.biolink_general)

    def test_biolink_model(self):
        self.rule_test(self.tests.biolinkmodel_entries)


if __name__ == '__main__':
    RewriteRuleTestCase.SERVER = os.environ.get('SERVER', RewriteRuleTestCase.SERVER)
    unittest.main()



import os
import unittest

# This has to occur post ClickTestCase
import click

from biolinkml.generators.jsonldcontextgen import cli
from tests import sourcedir, source_yaml_path
from tests.test_scripts.clicktestcase import ClickTestCase
from tests.utils.metadata_filters import ldcontext_metadata_filter


class GenContextTestCase(ClickTestCase):
    testdir = "gencontext"
    click_ep = cli
    prog_name = "gen-jsonld-context"

    def test_help(self):
        self.do_test("--help", 'help')

    def test_meta(self):
        self.maxDiff = None
        self.do_test(source_yaml_path, 'meta_context.jsonld', filtr=ldcontext_metadata_filter)
        self.do_test(source_yaml_path + ' --metauris', 'meta_contextn.jsonld', filtr=ldcontext_metadata_filter)
        self.do_test(source_yaml_path + ' -f xsv', 'meta_error', error=click.exceptions.BadParameter)
        self.do_test(source_yaml_path + ' --niggles', 'meta2_error', error=click.exceptions.NoSuchOption)

    def test_slot_class_uri(self):
        uri_tests_yaml = os.path.join(sourcedir, 'uri_tests.yaml')
        self.do_test(uri_tests_yaml, 'uri_tests.jsonld', filtr=ldcontext_metadata_filter)


if __name__ == '__main__':
    unittest.main()



import filecmp
import os.path
from contextlib import redirect_stdout
from io import StringIO
from typing import Optional


def are_dir_trees_equal(dir1: str, dir2: str) -> Optional[str]:
    """
    Compare two directories recursively. Files in each directory are
    assumed to be equal if their names and contents are equal.

    @param dir1: First directory path
    @param dir2: Second directory path

    @return: None if directories match, else summary of differences
   """

    dirs_cmp = filecmp.dircmp(dir1, dir2)
    output = StringIO()
    if dirs_cmp.diff_files or dirs_cmp.funny_files or dirs_cmp.left_only or dirs_cmp.right_only or dirs_cmp.funny_files:
        with redirect_stdout(output):
            dirs_cmp.report_full_closure()
        return output.getvalue()
    return None



import unittest

from biolinkml.utils.formatutils import camelcase, underscore, lcamelcase, be, split_line, wrapped_annotation


class FormatUtilsTestCase(unittest.TestCase):
    def test_formats(self):
        self.assertEqual("ThisIsIt", camelcase("this is it"))
        self.assertEqual("ThisIsIT", camelcase("  this   is iT   "))
        self.assertEqual("IBeY", camelcase("i be y "))
        self.assertEqual("ThisIsIt", camelcase("This__is_it"))

        self.assertEqual("this_is_it", underscore(" this is it "))
        self.assertEqual("this_is_it", underscore("this   is   it"))

        self.assertEqual("thisIsIt", lcamelcase("   this   is\t  it\n"))

        self.assertEqual('abc', be('  abc\n'))
        self.assertEqual('', be(None))
        self.assertEqual('', be('   '))

    def test_linestuff(self):
        text = "This is a mess'o test that goes on for a long way.  It has some carriage\n returns embedded in it " \
               "but otherwise it drags on and on and on until the cows come home.  Splitline covers this we hope."
        self.assertEqual(["This is a mess'o test that goes on for a long way. It has some carriage returns embedded"
                          " in it but otherwise it ",
                          'drags on and on and on until the cows come home. Splitline covers this we hope. '],
                         split_line(text))
        self.assertEqual(["This is a mess'o ", 'test that goes on ', 'for a long way. It ', 'has some carriage ',
                          'returns embedded in ', 'it but otherwise it ', 'drags on and on and ', 'on until the cows ',
                          'come home. ', 'Splitline covers ', 'this we hope. '], split_line(text, 20))
        self.assertEqual(['X' * 100 + ' '], split_line('X'*100, 20))
        self.assertEqual("""This is a mess'o test that goes on for a long way.  It has some carriage
	returns embedded in it but otherwise it drags on and on and on until the cows come home. Splitline covers this we 
	hope. """, wrapped_annotation(text))



if __name__ == '__main__':
    unittest.main()



import re
from typing import List

from pyshex.shex_evaluator import EvaluationResult

ws_pattern = re.compile(r'\s+')
us_pattern = re.compile(r'_+')


def camelcase(txt: str) -> str:
    def _up(s: str):
        return s[0].upper() + (s[1:] if len(s) > 1 else '')

    return ''.join([_up(word) for word in us_pattern.sub(' ', txt.strip().replace(',', '')).split()])


def underscore(txt: str) -> str:
    return ws_pattern.sub('_', txt.strip()).replace(',', '')


def lcamelcase(txt: str) -> str:
    s = camelcase(txt)
    return s[0].lower() + s[1:]


def be(entry: object) -> str:
    """ Return a stringified version of object replacing Nones with empty strings """
    return str(entry).strip() if entry else ''


split_col = 115


def sfx(uri: str) -> str:
    """
    Add a separator to a uri if none exists.

    Note: This should only be used with module id's -- it is not uncommon to use partial prefixes,
    e.g. PREFIX bfo: http://purl.obolibrary.org/obo/BFO_
    :param uri: uri to be suffixed
    :return: URI with suffix
    """
    return str(uri) + ('' if uri.endswith(('/', '#', '_', ':')) else '/')


def uri_for(prefix: str, suffix: str) -> str:
    """ Generator for predicate and identifier URI's """
    if ':' in prefix:
        return sfx(prefix) + suffix
    else:
        return prefix + ':' + suffix


def split_line(txt: str, split_len: int = split_col) -> List[str]:
    # TODO: consider replacing by textwrap.fill function, but note that its behavior is a bit different
    out_lines = []
    words = txt.split()
    cur_line = ""
    for word in words:
        word += ' '
        if len(cur_line) + len(word) > split_len:
            out_lines.append(cur_line if cur_line else word)
            if not cur_line:
                word = ""
            else:
                cur_line = ""
        cur_line += word
    if cur_line:
        out_lines.append(cur_line)
    return out_lines


def wrapped_annotation(txt: str) -> str:
    rval = []
    for line in [l.strip() for l in txt.split('\n')]:
        if len(line) > split_col:
            rval += split_line(line)
        else:
            rval.append(line)
    return '\n\t'.join(rval)


def shex_results_as_string(rslts: EvaluationResult) -> str:
    """ Pretty print ShEx Evaluation result """
    # TODO: Add this method to ShEx itself
    rval = [f"Evalutating: {str(rslts.focus)} against {str(rslts.start)}"]
    if rslts.result:
        rval.append("Result: CONFORMS")
    else:
        rval.append("Result: NonConforming")
    rval += rslts.reason.split('\n')
    return '\n'.join(rval)




# id: https://example.org/inheritedid
# description: Test
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace
from biolinkml.utils.metamodelcore import URI

metamodel_version = "1.4.3"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
BIOLINKML = CurieNamespace('biolinkml', 'https://w3id.org/biolink/biolinkml/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = BIOLINKML


# Types
class String(str):
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "string"
    type_model_uri = BIOLINKML.String


class Uri(URI):
    """ a complete URI """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uri"
    type_model_uri = BIOLINKML.Uri


class IdentifierType(String):
    """ A string that is intended to uniquely identify a thing May be URI in full or compact (CURIE) form """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "identifier type"
    type_model_uri = BIOLINKML.IdentifierType


class LabelType(String):
    """ A string that provides a human-readable name for a thing """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "label type"
    type_model_uri = BIOLINKML.LabelType


# Class references
class NamedThingId(IdentifierType):
    pass


class AttributeId(IdentifierType):
    pass


class BiologicalSexId(AttributeId):
    pass


class OntologyClassId(NamedThingId):
    pass


@dataclass
class NamedThing(YAMLRoot):
    """
    a databased entity or concept/class
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = BIOLINKML.NamedThing
    class_class_curie: ClassVar[str] = "biolinkml:NamedThing"
    class_name: ClassVar[str] = "named thing"
    class_model_uri: ClassVar[URIRef] = BIOLINKML.NamedThing

    id: Union[str, NamedThingId]
    name: Optional[Union[str, LabelType]] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.id is None:
            raise ValueError(f"id must be supplied")
        if not isinstance(self.id, NamedThingId):
            self.id = NamedThingId(self.id)
        if self.name is not None and not isinstance(self.name, LabelType):
            self.name = LabelType(self.name)
        super().__post_init__(**kwargs)


@dataclass
class Attribute(YAMLRoot):
    """
    A property or characteristic of an entity
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = BIOLINKML.Attribute
    class_class_curie: ClassVar[str] = "biolinkml:Attribute"
    class_name: ClassVar[str] = "attribute"
    class_model_uri: ClassVar[URIRef] = BIOLINKML.Attribute

    id: Union[str, AttributeId]

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.id is None:
            raise ValueError(f"id must be supplied")
        if not isinstance(self.id, AttributeId):
            self.id = AttributeId(self.id)
        super().__post_init__(**kwargs)


@dataclass
class BiologicalSex(Attribute):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = BIOLINKML.BiologicalSex
    class_class_curie: ClassVar[str] = "biolinkml:BiologicalSex"
    class_name: ClassVar[str] = "biological sex"
    class_model_uri: ClassVar[URIRef] = BIOLINKML.BiologicalSex

    id: Union[str, BiologicalSexId] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.id is None:
            raise ValueError(f"id must be supplied")
        if not isinstance(self.id, BiologicalSexId):
            self.id = BiologicalSexId(self.id)
        super().__post_init__(**kwargs)


@dataclass
class OntologyClass(NamedThing):
    """
    a concept or class in an ontology, vocabulary or thesaurus
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = BIOLINKML.OntologyClass
    class_class_curie: ClassVar[str] = "biolinkml:OntologyClass"
    class_name: ClassVar[str] = "ontology class"
    class_model_uri: ClassVar[URIRef] = BIOLINKML.OntologyClass

    id: Union[str, OntologyClassId] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.id is None:
            raise ValueError(f"id must be supplied")
        if not isinstance(self.id, OntologyClassId):
            self.id = OntologyClassId(self.id)
        super().__post_init__(**kwargs)



# Slots
class slots:
    pass

slots.id = Slot(uri=BIOLINKML.id, name="id", curie=BIOLINKML.curie('id'),
                      model_uri=BIOLINKML.id, domain=NamedThing, range=Union[str, NamedThingId])

slots.name = Slot(uri=BIOLINKML.name, name="name", curie=BIOLINKML.curie('name'),
                      model_uri=BIOLINKML.name, domain=NamedThing, range=Optional[Union[str, LabelType]])


from types import ModuleType

from jsonasobj import as_json, loads
from rdflib import Graph

from biolinkml.generators.jsonldcontextgen import ContextGenerator
from biolinkml.generators.pythongen import PythonGenerator
from biolinkml.generators.shexgen import ShExGenerator
from biolinkml.generators.yumlgen import YumlGenerator
from biolinkml.utils.yamlutils import as_json_object as yaml_to_json

yaml = """
id: http://example.org/sample/example1
name: synopsis2
prefixes:
    foaf: http://xmlns.com/foaf/0.1/
    ex: http://example.org/model/
    xsd: http://www.w3.org/2001/XMLSchema#
    
default_prefix: ex

default_curi_maps:
    - semweb_context
    
default_range: string

types:
    string:
        base: str
        uri: xsd:string
    int:
        base: int
        uri: xsd:integer
    boolean:
        base: Bool
        uri: xsd:boolean
        

classes:
    person:
        description: A person, living or dead
        slots:
            - id
            - first name
            - last name
            - age
            - living
            - knows
            
    friendly_person:
        description: Any person that knows someone
        is_a: person
        slot_usage:
            knows:
                required: True
            

slots:
    id:
        description: Unique identifier of a person
        identifier: true

    first name:
        description: The first name of a person
        slot_uri: foaf:firstName
        multivalued: true
        
    last name:
        description: The last name of a person
        slot_uri: foaf:lastName
        required: true
        
    living:
        description: Whether the person is alive
        range: boolean
        comments:
            - unspecified means unknown
        
    age:
        description: The age of a person if living or age of death if not
        range: int
        slot_uri: foaf:age
        
    knows:
        description: A person known by this person (indicating some level of reciprocated interaction between the parties).
        range: person
        slot_uri: foaf:knows
        multivalued: true
"""
python_src = PythonGenerator(yaml).serialize()
print(python_src)
spec = compile(PythonGenerator(yaml).serialize(), 'test', 'exec')
module = ModuleType('test')
exec(spec, module.__dict__)

print(f'<img src="{YumlGenerator(yaml).serialize()}"/>')
print(f'\n-----\n{YumlGenerator(yaml).serialize()}\n')

cntxt = loads(ContextGenerator(yaml).serialize(base="http://example.org/context/"))
print(as_json(cntxt))

shex = ShExGenerator(yaml).serialize(collections=False)
print(shex)


# Generate a person
joe_smith = module.Person(id="42", last_name="smith", first_name=['Joe', 'Bob'], age=43)
print(joe_smith)

# Add the context and turn it into RDF
jsonld = as_json(yaml_to_json(joe_smith, cntxt))
print(jsonld)
g = Graph()
g.parse(data=jsonld, format="json-ld")
print(g.serialize(format="turtle").decode())

from pyshex.evaluate import evaluate
r = evaluate(g, shex,
             start="http://example.org/sample/example1/Person",
             focus="http://example.org/context/42")
print("Conforms" if r[0] else r[1])



#!/usr/bin/env python
import sys
from setuptools import setup
from warnings import warn

if sys.version_info < (3, 7, 6):
    warn(f"Some URL processing will fail with python 3.7.5 or earlier.  Current version: {sys.version_info}")

setup(
    setup_requires=['pbr'],
    pbr=True,
)


import os
import unittest
from contextlib import redirect_stderr
from io import StringIO
from pprint import pprint
from typing import Union, TextIO, cast

from biolinkml import LOCAL_METAMODEL_YAML_FILE
from biolinkml.meta import SchemaDefinition, ClassDefinition, SlotDefinition, TypeDefinition, SubsetDefinition, \
    ElementName, SlotDefinitionName, ClassDefinitionName, TypeDefinitionName, Element
from biolinkml.utils.generator import Generator
from biolinkml.utils.typereferences import References
from tests.test_utils import inputdir


class GeneratorTest(Generator):
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ['txt']

    visit_all_class_slots: bool = True
    visits_are_sorted: bool = False
    sort_class_slots: bool = False

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], fmt: str = 'txt', emit_metadata: bool = False) \
            -> None:
        self.visited = []
        self.visit_class_return = True
        self.visit_all_class_slots: bool = True
        self.visits_are_sorted: bool = False
        self.sort_class_slots: bool = False
        super().__init__(schema, fmt, emit_metadata)

    def visit_schema(self, **kwargs) -> None:
        self.visited = ['init']
        self.visited.append(f'schema: {self.schema.name}')

    def end_schema(self, **kwargs) -> None:
        self.visited.append(f'end_schema: {self.schema.name}')
        pprint(self.visited)

    def visit_class(self, cls: ClassDefinition) -> bool:
        self.visited.append(f"class: {cls.name}")
        return self.visit_class_return

    def end_class(self, cls: ClassDefinition) -> None:
        self.visited.append(f"end_class: {cls.name}")

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> None:
        self.visited.append(f"  slot: {slot.name}" +
                            (f" ({aliased_slot_name})" if slot.name != aliased_slot_name else ''))

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        self.visited.append(f"slot: {slot.name}" +
                            (f" ({aliased_slot_name})" if slot.name != aliased_slot_name else ''))

    def visit_type(self, typ: TypeDefinition) -> None:
        self.visited.append(f"type: {typ.name}")

    def visit_subset(self, subset: SubsetDefinition) -> None:
        self.visited.append(f"subset: {subset.name}")


# visit_all_class_slots = True, visits_are_sorted = False, sort_class_slots = False
expected1 = [
     'init',
     'schema: generator1',
     'subset: ss2',
     'subset: ss1',
     'type: t2',
     'type: t1',
     'slot: slot1',
     'slot: slot2',
     'slot: mixin slot 1',
     'slot: mixin slot 2',
     'slot: mixin slot 3',
     'slot: applyto slot 1',
     'slot: applyto slot 2',
     'slot: c2_slot1 (slot1)',
     'slot: c3_applyto slot 2 (applyto slot 2)',
     'slot: c3_slot1 (slot1)',
     'class: c1',
     '  slot: slot1',
     '  slot: slot2',
     '  slot: mixin slot 1',
     '  slot: mixin slot 2',
     '  slot: mixin slot 3',
     'end_class: c1',
     'class: c2',
     '  slot: c2_slot1 (slot1)',
     '  slot: slot2',
     '  slot: mixin slot 1',
     '  slot: mixin slot 2',
     '  slot: mixin slot 3',
     '  slot: applyto slot 1',
     '  slot: applyto slot 2',
     'end_class: c2',
     'class: c3',
     '  slot: c3_slot1 (slot1)',
     '  slot: slot2',
     '  slot: mixin slot 1',
     '  slot: mixin slot 2',
     '  slot: mixin slot 3',
     '  slot: applyto slot 1',
     '  slot: c3_applyto slot 2 (applyto slot 2)',
     'end_class: c3',
     'class: c4',
     '  slot: applyto slot 1',
     '  slot: applyto slot 2',
     'end_class: c4',
     'class: mixin1',
     '  slot: mixin slot 1',
     '  slot: mixin slot 2',
     'end_class: mixin1',
     'class: mixin2',
     '  slot: mixin slot 3',
     'end_class: mixin2',
     'class: applyto1',
     '  slot: applyto slot 1',
     '  slot: applyto slot 2',
     'end_class: applyto1',
     'end_schema: generator1']

# visit_all_class_slots = False, visits_are_sorted = False, sort_class_slots = False
expected2 = [
    'init',
    'schema: generator1',
    'subset: ss2',
    'subset: ss1',
    'type: t2',
    'type: t1',
    'slot: slot1',
    'slot: slot2',
    'slot: mixin slot 1',
    'slot: mixin slot 2',
    'slot: mixin slot 3',
    'slot: applyto slot 1',
    'slot: applyto slot 2',
    'slot: c2_slot1 (slot1)',
    'slot: c3_applyto slot 2 (applyto slot 2)',
    'slot: c3_slot1 (slot1)',
    'class: c1',
    '  slot: slot1',
    '  slot: slot2',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    'end_class: c1',
    'class: c2',
    '  slot: c2_slot1 (slot1)',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: c2',
    'class: c3',
    '  slot: c3_slot1 (slot1)',
    '  slot: c3_applyto slot 2 (applyto slot 2)',
    'end_class: c3',
    'class: c4',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: c4',
    'class: mixin1',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    'end_class: mixin1',
    'class: mixin2',
    '  slot: mixin slot 3',
    'end_class: mixin2',
    'class: applyto1',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: applyto1',
    'end_schema: generator1']

# visit_all_class_slots = True, visits_are_sorted = True, sort_class_slots = False
expected3 = [
    'init',
    'schema: generator1',
    'subset: ss1',
    'subset: ss2',
    'type: t1',
    'type: t2',
    'slot: applyto slot 1',
    'slot: applyto slot 2',
    'slot: c2_slot1 (slot1)',
    'slot: c3_applyto slot 2 (applyto slot 2)',
    'slot: c3_slot1 (slot1)',
    'slot: mixin slot 1',
    'slot: mixin slot 2',
    'slot: mixin slot 3',
    'slot: slot1',
    'slot: slot2',
    'class: applyto1',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: applyto1',
    'class: c1',
    '  slot: slot1',
    '  slot: slot2',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    'end_class: c1',
    'class: c2',
    '  slot: c2_slot1 (slot1)',
    '  slot: slot2',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: c2',
    'class: c3',
    '  slot: c3_slot1 (slot1)',
    '  slot: slot2',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    '  slot: applyto slot 1',
    '  slot: c3_applyto slot 2 (applyto slot 2)',
    'end_class: c3',
    'class: c4',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: c4',
    'class: mixin1',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    'end_class: mixin1',
    'class: mixin2',
    '  slot: mixin slot 3',
    'end_class: mixin2',
    'end_schema: generator1']

# visit_all_class_slots = True, visits_are_sorted = False, sort_class_slots = True
expected4 = [
    'init',
    'schema: generator1',
    'subset: ss2',
    'subset: ss1',
    'type: t2',
    'type: t1',
    'slot: slot1',
    'slot: slot2',
    'slot: mixin slot 1',
    'slot: mixin slot 2',
    'slot: mixin slot 3',
    'slot: applyto slot 1',
    'slot: applyto slot 2',
    'slot: c2_slot1 (slot1)',
    'slot: c3_applyto slot 2 (applyto slot 2)',
    'slot: c3_slot1 (slot1)',
    'class: c1',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    '  slot: slot1',
    '  slot: slot2',
    'end_class: c1',
    'class: c2',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    '  slot: c2_slot1 (slot1)',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    '  slot: slot2',
    'end_class: c2',
    'class: c3',
    '  slot: applyto slot 1',
    '  slot: c3_applyto slot 2 (applyto slot 2)',
    '  slot: c3_slot1 (slot1)',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    '  slot: mixin slot 3',
    '  slot: slot2',
    'end_class: c3',
    'class: c4',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: c4',
    'class: mixin1',
    '  slot: mixin slot 1',
    '  slot: mixin slot 2',
    'end_class: mixin1',
    'class: mixin2',
    '  slot: mixin slot 3',
    'end_class: mixin2',
    'class: applyto1',
    '  slot: applyto slot 1',
    '  slot: applyto slot 2',
    'end_class: applyto1',
    'end_schema: generator1']

expected5 = [
    'init',
    'schema: generator1',
    'subset: ss2',
    'subset: ss1',
    'type: t2',
    'type: t1',
    'slot: slot1',
    'slot: slot2',
    'slot: mixin slot 1',
    'slot: mixin slot 2',
    'slot: mixin slot 3',
    'slot: applyto slot 1',
    'slot: applyto slot 2',
    'slot: c2_slot1 (slot1)',
    'slot: c3_applyto slot 2 (applyto slot 2)',
    'slot: c3_slot1 (slot1)',
    'class: c1',
    'class: c2',
    'class: c3',
    'class: c4',
    'class: mixin1',
    'class: mixin2',
    'class: applyto1',
    'end_schema: generator1']


class GeneratorTestCase(unittest.TestCase):
    def test_visitors(self):
        """ Test the generator visitor functions """
        gen = GeneratorTest(os.path.join(inputdir, 'generator1.yaml'))
        gen.serialize()
        self.assertEqual(expected1, gen.visited)
        gen.visit_all_class_slots = False
        gen.serialize()
        self.assertEqual(expected2, gen.visited)
        gen.visit_all_class_slots = True
        gen.visits_are_sorted = True
        gen.serialize()
        self.assertEqual(expected3, gen.visited)
        gen.visits_are_sorted = False
        gen.sort_class_slots = True
        gen.serialize()
        self.assertEqual(expected4, gen.visited)
        gen.sort_class_slots = False
        gen.visit_class_return = False
        gen.serialize()
        self.assertEqual(expected5, gen.visited)

    def test_default_prefix(self):
        """ Test default prefix utility """
        model = """
id: http://example.org/test/t1
name: t1

default_range: string
types:
    string:
        base: str
        uri: xsd:string

prefixes:
    xsd: http://www.w3.org/2001/XMLSchema#
    AAA: http://example.org/test/aaa/
    BBB: http://example.org/test/bbb/
"""
        gen = GeneratorTest(model + "\n\ndefault_prefix: AAA")
        self.assertEqual("http://example.org/test/aaa/", gen.default_prefix())

        gen = GeneratorTest(model + "\n\ndefault_prefix: http://example.org/test/default/")
        self.assertEqual("http://example.org/test/default/", gen.default_prefix())

        with self.assertRaises(ValueError):
            GeneratorTest(model + "\n\ndefault_prefix: CCCC")

    def test_duplicate_names(self):
        """ Test duplicate name for slot and type detection """
        model = """
id: http://example.org/test/t1
name: t1
default_range: string

types:
    string:
        base: str
        
    dup name:
        base: int
        
    dn2:

        
classes:
    dn2:
    
    dup name:
"""
        with self.assertRaises(ValueError):
            GeneratorTest(model)

    def test_element_name(self):
        """ Test formatted_element_name """
        model = """
id: http://example.org/test/t1
name: t1

prefixes:
    xsd: http://www.w3.org/2001/XMLSchema#

subsets:
    dup name:
    subset ss1:

default_range: dup name

types:
    str:
        base: str
        uri: xsd:string
    dup name:
        base: int
        uri: xsd:integer
    type t1:
        base: int
        uri: xsd:integer
    
        
slots:
    dup name:
        domain: class c1
        
classes:
    class c1:
"""
        errfile = StringIO()
        with redirect_stderr(errfile):
            gen = GeneratorTest(model)

        # Note: There is something about the PyCharm / UnitTest package that, if you are running a lot of tests, the
        # output ends up getting redirected to the test runner rather than stderr.  If there is nothing at all, we
        # will assume that this is the case.
        if errfile.getvalue().strip():
            self.assertEqual("""Warning: Shared type and slot names: dup name
Warning: Shared slot and subset names: dup name
Warning: Shared type and subset names: dup name""", errfile.getvalue().strip())
        else:
            print("*** Warning not tested -- stderr redirect isn't working")


        self.assertEqual('dup_name', gen.formatted_element_name(cast(ElementName, 'dup name'), False))
        self.assertEqual('int', gen.formatted_element_name(cast(ElementName, 'dup name'), True))
        self.assertEqual('str', gen.formatted_element_name(cast(ElementName, 'str'), True))
        self.assertIsNone(gen.formatted_element_name(cast(ElementName, 'class c1'), False))
        self.assertEqual('ClassC1', gen.formatted_element_name(cast(ElementName, 'class c1'), True))
        self.assertEqual('SubsetSs1', gen.formatted_element_name(cast(ElementName, 'subset ss1'), False))
        self.assertEqual("Unknown_ClassC2", gen.class_or_type_name(cast(ElementName, 'class c2')))
        self.assertEqual("int", gen.class_or_type_name(cast(ElementName, 'dup name')))
        self.assertEqual("int", gen.class_or_type_name(cast(ElementName, 'type t1')))
        self.assertEqual("unknown_slot_s2", gen.slot_name(cast(SlotDefinitionName, 'slot s2')))
        self.assertEqual('Unknown_SS', gen.subset_name(cast(ElementName, 's s')))

        self.assertEqual('dup_name', gen.formatted_element_name(gen.schema.slots[cast(ElementName, 'dup name')]))
        self.assertEqual('int', gen.formatted_element_name(gen.schema.types[cast(ElementName, 'dup name')]))
        self.assertEqual('DupName', gen.formatted_element_name(gen.schema.subsets[cast(ElementName, 'dup name')]))
        self.assertEqual('ClassC1', gen.formatted_element_name(gen.schema.classes[cast(ElementName, 'class c1')]))
        self.assertIsNone(gen.formatted_element_name(cast(Element, gen)))

    def test_own_slots(self):
        """ Test the generator own_slots and all_slots helper functions """
        gen = GeneratorTest(os.path.join(inputdir, 'ownalltest.yaml'))
        gen.sort_class_slots = True

        self.assertEqual(['s6'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'at1'))])
        self.assertEqual(['s6'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'at1'))])
        self.assertEqual(['s6'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'at1'),
                                                                cls_slots_first=True)])

        self.assertEqual(['s5', 's6'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'm2'))])
        self.assertEqual(['s5', 's6'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'm2'))])
        self.assertEqual(['s5', 's6'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'm2'),
                                                                      cls_slots_first=True)])

        self.assertEqual(['s4'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'm1'))])
        self.assertEqual(['s4'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'm1'))])
        self.assertEqual(['s4'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'm1'),
                                                                cls_slots_first=True)])

        self.assertEqual(['s1', 's3'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'c1'))])
        self.assertEqual(['s1', 's3'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c1'))])
        self.assertEqual(['s1', 's3'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c1'),
                                                                      cls_slots_first=True)])

        self.assertEqual(['s2', 's4'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'c2'))])
        self.assertEqual(['s1', 's2', 's3', 's4'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c2'))])
        self.assertEqual(['s2', 's4', 's1', 's3'], [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c2'),
                                                                                  cls_slots_first=True)])

        self.assertEqual(['s5', 's6'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'c3'))])
        self.assertEqual(['s1', 's2', 's3', 's4', 's5', 's6'], [s.name for s in gen.all_slots(cast(ClassDefinitionName,
                                                                                                   'c3'))])
        self.assertEqual(['s5', 's6', 's2', 's4', 's1', 's3'],
                         [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c3'), cls_slots_first=True)])
        
        self.assertEqual(['c4_s1', 'c4_s5', 'c4_s6'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'c4'))])
        self.assertEqual(['c4_s1', 'c4_s5', 'c4_s6', 's2', 's3', 's4'],
                         [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c4'))])
        self.assertEqual(['c4_s1', 'c4_s5', 'c4_s6', 's2', 's4', 's3'],
                         [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c4'), cls_slots_first=True)])
        
        self.assertEqual(['c5_s1', 'c5_s6'], [s.name for s in gen.own_slots(cast(ClassDefinitionName, 'c5'))])
        self.assertEqual(['c4_s5', 'c5_s1', 'c5_s6', 's2', 's3', 's4'],
                         [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c5'))])
        self.assertEqual(['c5_s1', 'c5_s6', 'c4_s5', 's2', 's4', 's3'],
                         [s.name for s in gen.all_slots(cast(ClassDefinitionName, 'c5'), cls_slots_first=True)])

    def test_slot_class_paths(self):
        """ Test for aliased slot name, class identifier path and slot type path """
        gen = GeneratorTest(os.path.join(inputdir, 'ownalltest.yaml'))
        gen.sort_class_slots = True
        self.assertEqual(['s1', 's5', 's6', 's2', 's3', 's4'],
                         [gen.aliased_slot_name(s.name) for s in gen.all_slots(cast(ClassDefinitionName, 'c4'))])
        self.assertEqual(['s5', 's1', 's6', 's2', 's3', 's4'],
                         [gen.aliased_slot_name(s) for s in gen.all_slots(cast(ClassDefinitionName, 'c5'))])
        self.assertEqual({
            'c4_s1': ['int'],
            'c4_s5': ['Bool', 'T5'],
            'c4_s6': ['int', 'C1S1', 'C2S1', 'C3S1', 'C4S1'],
            's2': ['int', 'C1S1'],
            's3': ['int', 'T2', 'T3'],
            's4': ['Bool']}, {s.name: gen.slot_type_path(s) for s in gen.all_slots(cast(ClassDefinitionName, 'c4'))})
        self.assertEqual({
            'c4_s5': ['Bool', 'T5'],
            'c5_s1': ['int'],
            'c5_s6': ['str'],
            's2': ['int', 'C1S1'],
            's3': ['int', 'T2', 'T3'],
            's4': ['Bool']}, {s.name: gen.slot_type_path(s) for s in gen.all_slots(cast(ClassDefinitionName, 'c5'))})
        self.assertEqual({'s1': ['int'], 's3': ['int', 'T2', 'T3']},
                         {s.name: gen.slot_type_path(s) for s in gen.all_slots(cast(ClassDefinitionName, 'c1'))})

    def test_ancestors(self):
        """ Test ancestors function and duplicate name detection """
        model = """
id: http://example.org/test/t1
name: t1

prefixes:
  xsd: http://www.w3.org/2001/XMLSchema#

default_range: string

types:
    string:
        base: str
        uri: xsd:string

slots:
    slot s1:
        domain: class c2

    slot s2:
        is_a: slot s1
        
    slot s3:
        is_a: slot s2
        
    slot s4:
        is_a: slot s2
        
classes:
    slot s1:
    
    class c2:
        is_a: slot s1
"""
        errfile = StringIO()
        with redirect_stderr(errfile):
            gen = GeneratorTest(model)
        # Note: There is something about the PyCharm / UnitTest package that, if you are running a lot of tests, the
        # output ends up getting redirected to the test runner rather than stderr.  If there is nothing at all, we
        # will assume that this is the case.
        if errfile.getvalue().strip():
            self.assertEqual("WARNING:GeneratorTest:Shared class and slot names: slot s1", errfile.getvalue().strip())
        else:
            print("*** Warning not tested -- stderr redirect isn't working")

        self.assertEqual(['slot s1'], gen.ancestors(gen.schema.slots['slot s1']))
        self.assertEqual(['slot s2', 'slot s1'], gen.ancestors(gen.schema.slots[cast(ElementName, 'slot s2')]))
        self.assertEqual(['slot s3', 'slot s2', 'slot s1'], gen.ancestors(gen.schema.slots['slot s3']))
        self.assertEqual(['slot s4', 'slot s2', 'slot s1'], gen.ancestors(gen.schema.slots['slot s4']))
        self.assertEqual(['slot s1'], gen.ancestors(gen.schema.classes['slot s1']))
        self.assertEqual(['class c2', 'slot s1'], gen.ancestors(gen.schema.classes['class c2']))

    def test_range_type_path(self):
        model = """
id: http://example.org/test/t1
name: t1

prefixes:
  xsd: http://www.w3.org/2001/XMLSchema#

default_range: type 1
types:
    type 1:
        base: int
        uri: xsd:integer
        
    type 2:
        typeof: type 1
        
    type 3:
        typeof: type 2
"""
        gen = GeneratorTest(model)
        self.assertEqual(['int'], gen.range_type_path(gen.schema.types['type 1']))
        self.assertEqual(['int', 'Type2'], gen.range_type_path(gen.schema.types['type 2']))
        self.assertEqual(['int', 'Type2', 'Type3'], gen.range_type_path(gen.schema.types['type 3']))

    def test_identifier_path(self):
        """ Test the case of an implicitly inlined class """
        model = """
id: http://example.org/test/t1
name: t1

prefixes:
    xsd: http://www.w3.org/2001/XMLSchema#
    
types:
    string:
        base: str
        uri: xsd:string
        
slots:
    s1:
        domain: c1
        range: c2
        
    s2:
        domain: c2
        range: string
        
classes:
    c1:
    c2:

"""
        gen = GeneratorTest(model)
        self.assertEqual({'s1': ['dict', 'C2'], 's2': ['str']},
                         {s.name: gen.slot_type_path(s) for s in gen.schema.slots.values()})

    def test_meta_neighborhood(self):
        """ Test the neighborhood function in the metamodel """
        gen = GeneratorTest(LOCAL_METAMODEL_YAML_FILE)
        errfile = StringIO()
        with redirect_stderr(errfile):
            gen.neighborhood([cast(ElementName, 'Definition')])
            # Note: There is something about the PyCharm / UnitTest package that, if you are running a lot of tests, the
            # output ends up getting redirected to the test runner rather than stderr.  If there is nothing at all, we
            # will assume that this is the case.
            if errfile.getvalue().strip():
                self.assertEqual("Warning: neighborhood(Definition) - Definition is undefined",
                                 errfile.getvalue().strip())
            else:
                print("*** Warning not tested -- stderr redirect isn't working")
        self.assertEqual(References(
            classrefs={cast(ClassDefinitionName, e) for e in ['class_definition', 'local_name',
                                                              'subset_definition', 'alt_description', 'definition',
                                                              'example', 'slot_definition', 'element']},
            slotrefs={cast(SlotDefinitionName, e) for e in ['is_a', 'apply_to', 'mixins', 'owner']},
            typerefs={cast(TypeDefinitionName, e) for e in ['uriorcurie', 'boolean', 'uri', 'ncname', 'string']},
            subsetrefs=set()), gen.neighborhood('definition'))


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from biolinkml.generators.pythongen import PythonGenerator
from tests.test_issues import sourcedir, outputdir


class Issue44UnitTest(unittest.TestCase):
    def test_multiple_postinit(self):
        """ Generate postinit code for a multi-occurring element """
        python = PythonGenerator(os.path.join(sourcedir, 'issue_44.yaml'), emit_metadata=False).serialize()
        os.makedirs(outputdir, exist_ok=True)
        outfile = os.path.join(outputdir, 'issue_44.py')
        if not os.path.exists(outfile):
            with open(outfile, 'w') as f:
                f.write(python)
            self.fail(f"Writing {outfile} - rerun test")
        else:
            with open(outfile) as f:
                old_python = f.read()
        self.assertEqual(old_python, python, f"Remove {outfile} to update target")


if __name__ == '__main__':
    unittest.main()



import unittest

# This has to occur post ClickTestCase
import click
from biolinkml.generators.jsonschemagen import cli
from tests import source_yaml_path
from tests.test_scripts.clicktestcase import ClickTestCase


class GenJSONSchemaTestCase(ClickTestCase):
    testdir = "genjsonschema"
    click_ep = cli
    prog_name = "gen-json-schema"

    def test_help(self):
        self.do_test("--help", 'help', tox_wrap_fix=True)

    def test_meta(self):
        self.maxDiff = None
        self.do_test(source_yaml_path, 'meta.jsonld')
        self.do_test(source_yaml_path + ' -f json', 'meta.jsonld')
        self.do_test(source_yaml_path + ' -f xsv', 'meta_error', error=click.exceptions.BadParameter)
        self.do_test(source_yaml_path + " -i", 'meta_inline.json')


if __name__ == '__main__':
    unittest.main()



# Auto generated from .yaml by pythongen.py version:
# Generation date:
# Schema: testMetamodelMappings
#
# id: http://example.org/mappings/
# description: Mappings test
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from biolinkml.utils.slot import Slot
from biolinkml.utils.metamodelcore import empty_list, empty_dict, bnode
from biolinkml.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
if sys.version_info < (3, 7, 6):
    from biolinkml.utils.dataclass_extensions_375 import dataclasses_init_fn_with_kwargs
else:
    from biolinkml.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from biolinkml.utils.formatutils import camelcase, underscore, sfx
from rdflib import Namespace, URIRef
from biolinkml.utils.curienamespace import CurieNamespace


metamodel_version = "1.4.3"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
EX = CurieNamespace('ex', 'http://example.org/mappings/')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = EX


# Types
class String(str):
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "string"
    type_model_uri = EX.String


class IdentifierType(String):
    """ A string that is intended to uniquely identify a thing May be URI in full or compact (CURIE) form """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "identifier type"
    type_model_uri = EX.IdentifierType


# Class references
class C1S2(IdentifierType):
    pass


class C2S2(C1S2):
    pass


class C3S2(C1S2):
    pass


@dataclass
class C1(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = EX.C1
    class_class_curie: ClassVar[str] = "ex:C1"
    class_name: ClassVar[str] = "c1"
    class_model_uri: ClassVar[URIRef] = EX.C1

    s2: Union[str, C1S2]
    s1: Optional[str] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.s2 is None:
            raise ValueError(f"s2 must be supplied")
        if not isinstance(self.s2, C1S2):
            self.s2 = C1S2(self.s2)
        super().__post_init__(**kwargs)


@dataclass
class C2(C1):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = EX.C2
    class_class_curie: ClassVar[str] = "ex:C2"
    class_name: ClassVar[str] = "c2"
    class_model_uri: ClassVar[URIRef] = EX.C2

    s2: Union[str, C2S2] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.s2 is None:
            raise ValueError(f"s2 must be supplied")
        if not isinstance(self.s2, C2S2):
            self.s2 = C2S2(self.s2)
        super().__post_init__(**kwargs)


@dataclass
class C3(C1):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = EX.C3
    class_class_curie: ClassVar[str] = "ex:C3"
    class_name: ClassVar[str] = "c3"
    class_model_uri: ClassVar[URIRef] = EX.C3

    s2: Union[str, C3S2] = None

    def __post_init__(self, **kwargs: Dict[str, Any]):
        if self.s2 is None:
            raise ValueError(f"s2 must be supplied")
        if not isinstance(self.s2, C3S2):
            self.s2 = C3S2(self.s2)
        super().__post_init__(**kwargs)



# Slots
class slots:
    pass

slots.s1 = Slot(uri=EX.s1, name="s1", curie=EX.curie('s1'),
                      model_uri=EX.s1, domain=C1, range=Optional[str])

slots.s2 = Slot(uri=EX.s2, name="s2", curie=EX.curie('s2'),
                      model_uri=EX.s2, domain=C1, range=Union[str, C1S2])


from logging import ERROR

model = """
# Every model must have a globally unique URI. This is the external name of the model
id: http://example.org/examples/distributeExample 

# Every model should also have a globally unique name (well, global within the context of the particular modeling environment)
name: dist1

# Descriptions are always useful, but not required
description: A toy extension to the base biolink model

# Versions are recommended but not required.  The version is copied into the output artifacts.  An error will be raised
# if two different versions of the same model are imported
version: 0.0.1

# A license is not required at this point -- should it be?
license: https://creativecommons.org/publicdomain/zero/1.0/
                                                         
# Prefixes can be assigned specifically.  We define two below:
#   biolink -- the prefix used by the biolink-model
#   dist1 -- the URI prefix used by this example.  Note that the dist1 prefix may or may not align with the model id
prefixes:
  biolink: https://w3id.org/biolink/vocab/
  biolinkml: https://w3id.org/biolink/biolinkml/
  dist: http://example.org/examples/dist1#

# Prefixes can also be pulled from a prefixcommons compliant site. The map below uses the definitions found in
#     https://github.com/prefixcommons/biocontext/blob/master/registry/semweb_context.yaml.
default_curi_maps:
  - semweb_context

# The default prefix is what is used in the subsets, types, slots, classes sections below if not otherwise supplied
default_prefix:  dist
default_range: string

# The list of prefixes to emit target files.  Note that all prefixes that are used elsewhere in the model are automatically
# emitted, with the exception of 
emit_prefixes:
    - skos
    - rdf
    - dist

# List of models to import.  Note that import specifications can (currently) be URI's, absolute (file://...file), curies
# (biolink:model), or relative (includes/myfile) file names.  Note, however, that this latter form is being deprecated.  
# The location of imported files can now be specified in an accompanying mapping file.  The imports below reference:
#   https://w3id.org/biolink/biolink-model    -- the biolink model
#   https://w3id.org/biolink/biolinkml/types  -- the biolink modeling language types definitions
imports: 
  - https://w3id.org/biolink/biolink-model
  - biolinkml:types

# Subsets that are defined in this model extension
subsets:
    experimental:
        # A subset should have a description
        description: model elements that have not yet been tested

# Types that are defined in this model extension
types:
  gene sequence:
    uri: dist:seq
    typeof: string
    description: A gene sequence
        
# Slots that are defined in this model extension
slots:
    gene has sequence:
        description: A gene pattern
        domain: gene
        range: gene sequence
        slot_uri: dist:hasSeq
        required: true

# Classes that are defined in this model extension
classes:
    # The class name.  For most generators, this will be transformed to CamelCase (MyGene)
    my gene:
        description: This is an example extension.  Doesn't do a lot
        is_a: gene
        slots:
            - gene has sequence
    
    
"""


from biolinkml.generators.pythongen import PythonGenerator

gen = PythonGenerator(model, log_level=ERROR)
print(gen.serialize())




import os
import unittest
from types import ModuleType

from biolinkml import INCLUDES_DIR, MODULE_DIR
from biolinkml.generators.pythongen import PythonGenerator
from tests import targetdir
from tests.utils.generator_utils import GeneratorTestCase, BIOLINK_IMPORT_MAP
from tests.utils.metadata_filters import metadata_filter


class PythonTestCase(GeneratorTestCase):
    """ Generate python for all of the models, compare them against what has been published
     and verify that they compile"""

    target_path = targetdir

    def check_python(self) -> None:
        # Make sure the python is valid
        with open(os.path.join(self.source_path, self.model_name + '.py')) as f:
            pydata = f.read()
        spec = compile(pydata, 'test', 'exec')
        module = ModuleType('test')
        exec(spec, module.__dict__)

    def test_types_python(self):
        self.source_path = INCLUDES_DIR
        self.model_path = INCLUDES_DIR
        self.model_name = 'types'
        self.single_file_generator('py', PythonGenerator, filtr=metadata_filter)
        self.check_python()

    def test_mappings_python(self):
        self.source_path = INCLUDES_DIR
        self.model_path = INCLUDES_DIR
        self.model_name = 'mappings'
        self.importmap = BIOLINK_IMPORT_MAP
        self.single_file_generator('py', PythonGenerator, filtr=metadata_filter)
        self.check_python()

    def test_metamodel_python(self):
        self.source_path = os.path.join(MODULE_DIR, 'biolinkml')
        self.model_path = MODULE_DIR
        self.model_name = 'meta'
        self.importmap = BIOLINK_IMPORT_MAP
        self.single_file_generator('py', PythonGenerator, filtr=metadata_filter)
        self.check_python()


if __name__ == '__main__':
    unittest.main()



import os
import unittest

# This has to occur post ClickTestCase
import click

from biolinkml.generators.dotgen import cli
from tests import source_yaml_path
from tests.test_scripts.clicktestcase import ClickTestCase


class GraphvizTestCase(ClickTestCase):
    testdir = "gengraphviz"
    click_ep = cli
    prog_name = "gen-graphviz"

    def test_help(self):
        self.do_test("--help", 'help')

    @unittest.skipIf(True, 'Determine whether we need graphviz before moving further')
    def test_meta(self):

        # ALL may be useful, but it is very time consuming
        # NOTE: Because the GraphViz output binaries can vary considerably, we don't try to compare them
        outdir = self.temp_directory('meta')
        self.do_test(source_yaml_path + f" -d {outdir} -o all", dirbase='meta', comparator=self.always_pass_comparator)
        outdir = self.temp_directory('meta1')
        self.do_test(source_yaml_path + f" -f svg -d {outdir} -o all", dirbase='meta1',
                     comparator=self.always_pass_comparator)
        self.do_test(source_yaml_path + f' -f xyz -d {outdir} -o all', error=click.exceptions.BadParameter)
        outdir = os.path.join(self.tmpdir_path, 'meta2')
        self.do_test(source_yaml_path + f" -d {outdir} -c definition", dirbase='meta2',
                     comparator=self.always_pass_comparator)
        outdir = os.path.join(self.tmpdir_path, 'meta3')
        self.do_test([source_yaml_path, "-d", outdir, "-c", "class_definition", "-c", "element"], dirbase='meta3',
                     comparator=self.always_pass_comparator)
        self.do_test([source_yaml_path, "-c", "nada"], error=ValueError)

if __name__ == '__main__':
    unittest.main()



"""Generate OWL ontology corresponding to information model

model classes are translated to OWL classes, slots to OWL properties.
"""
import os
from typing import Union, TextIO, Optional, cast

import click
from rdflib import Graph, URIRef, RDF, OWL, Literal, BNode
from rdflib.collection import Collection
from rdflib.namespace import RDFS
from rdflib.plugin import plugins as rdflib_plugins, Parser as rdflib_Parser

from biolinkml import LOCAL_METAMODEL_YAML_FILE, METAMODEL_NAMESPACE_NAME, METAMODEL_NAMESPACE, METAMODEL_YAML_URI, META_BASE_URI
from biolinkml.meta import ClassDefinitionName, SchemaDefinition, ClassDefinition, SlotDefinitionName, \
    TypeDefinitionName, SlotDefinition, TypeDefinition, Element
from biolinkml.utils.formatutils import camelcase, underscore
from biolinkml.utils.generator import Generator, shared_arguments
from biolinkml.utils.schemaloader import SchemaLoader


class ElementDefinition(object):
    pass


class OwlSchemaGenerator(Generator):
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ['ttl'] + [x.name for x in rdflib_plugins(None, rdflib_Parser) if '/' not in str(x.name)]
    visits_are_sorted = True

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], **kwargs) -> None:
        super().__init__(schema, **kwargs)
        self.graph: Optional[Graph] = None
        self.metamodel = SchemaLoader(LOCAL_METAMODEL_YAML_FILE) if os.path.exists(LOCAL_METAMODEL_YAML_FILE) else\
            SchemaLoader(METAMODEL_YAML_URI, base_dir=META_BASE_URI)
        self.metamodel.resolve()
        self.top_value_uri: Optional[URIRef] = None

    def visit_schema(self, output: Optional[str] = None, **_):
        base = URIRef(self.schema.id)
        self.graph = Graph(identifier=base)
        for prefix in self.metamodel.schema.emit_prefixes:
            self.graph.bind(prefix, self.metamodel.namespaces[prefix])

        self.graph.add((base, RDF.type, OWL.Ontology))
        self._add_element_properties(base, self.schema)

        # add the model types
        for name in ['class_definition', 'type_definition', 'slot_definition', 'subset_definition']:
            self._add_metamodel_class(name)

        # add value placeholder
        self.top_value_uri = self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME]['topValue']
        self.graph.add((self.top_value_uri, RDF.type, OWL.DatatypeProperty))
        self.graph.add((self.top_value_uri, RDFS.label, Literal("value")))

    def end_schema(self, output: Optional[str] = None, **_) -> None:
        data = self.graph.serialize(format='turtle' if self.format == 'ttl' else self.format).decode()
        if output:
            with open(output, 'w') as outf:
                outf.write(data)
        else:
            print(data)

    def visit_class(self, cls: ClassDefinition) -> bool:
        cls_uri = self._class_uri(cls.name)
        self.graph.add((cls_uri, RDF.type, OWL.Class))
        self.graph.add((cls_uri, RDF.type,
                        self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][camelcase('class definition')]))
        self._add_element_properties(cls_uri, cls)

        # Parent classes
        # TODO: reintroduce this
        # if not cls.defining_slots:
        if True:
            if cls.is_a:
                self.graph.add((cls_uri, RDFS.subClassOf, self._class_uri(cls.is_a)))
            if cls.mixin:
                self.graph.add((cls_uri, RDFS.subClassOf, METAMODEL_NAMESPACE.mixin))
            for mixin in sorted(cls.mixins):
                self.graph.add((cls_uri, RDFS.subClassOf, self._class_uri(mixin)))
            if cls.name in self.synopsis.applytorefs:
                for appl in sorted(self.synopsis.applytorefs[cls.name].classrefs):
                    self.graph.add((cls_uri, RDFS.subClassOf, self._class_uri(appl)))
        else:
            raise NotImplementedError("Defining slots need to be implemented")
            # If defining slots, we generate an equivalentClass entry
            # equ_node = BNode()
            # self.graph.add((cls_uri, OWL.equivalentClass, equ_node))
            # self.graph.add((equ_node, RDF.type, OWL.Class))
            #
            # elts = []
            # if cls.is_a:
            #     elts.append(self._class_uri(cls.is_a))
            # if cls.mixin:
            #     self.graph.add((cls_uri, RDFS.subClassOf, META_NS.mixin))
            # for mixin in cls.mixins:
            #     self.graph.add((cls_uri, RDFS.subClassOf, self._class_uri(mixin)))
            # if cls.name in self.synopsis.applytorefs:
            #     for appl in self.synopsis.applytorefs[cls.name].classrefs:
            #         self.graph.add((cls_uri, RDFS.subClassOf, self._class_uri(appl)))
            #
            # for slotname in cls.defining_slots:
            #     restr_node = BNode()
            #     slot = self.schema.slots[slotname]
            #
            #     self.graph.add((restr_node, RDF.type, OWL.Restriction))
            #     self.graph.add((restr_node, OWL.onProperty, self._prop_uri(slotname)))
            #     self._add_cardinality(restr_node, slot)
            #     # TODO: fix this
            #     # self.graph.add((restr_node, OWL.someValuesFrom, self._build_range(slot)))
            #     elts.append(restr_node)
            #
            # coll_bnode = BNode()
            # Collection(self.graph, coll_bnode, elts)
            # self.graph.add((equ_node, OWL.intersectionOf, coll_bnode))

        # TODO: see whether unions belong
        # if cls.union_of:
        #     union_node = BNode()
        #     Collection(self.graph, union_coll, [self.class_uri(union_node) for union_node in cls.union_of])
        #     self.graph.add((union_node, OWL.unionOf, union_coll))
        #     self.graph.add((cls_uri, RDFS.subClassOf, union_node))

        for sn in sorted(self.own_slot_names(cls)):
            # Defining_slots are covered above
            if sn not in cls.defining_slots:
                slot = self.schema.slots[sn]
                # Non-inherited slots are annotation properties
                if self.is_slot_object_property(slot):
                    slot_node = BNode()
                    self.graph.add((cls_uri, RDFS.subClassOf, slot_node))

                    #         required multivalued
                    if slot.required:
                        if slot.multivalued:
                            #    y         y     intersectionOf(restriction(slot only type) restriction(slot some type)
                            restr1 = BNode()
                            self.graph.add((restr1, RDF.type, OWL.Restriction))
                            self.graph.add((restr1, OWL.allValuesFrom, self._range_uri(slot)))
                            self.graph.add((restr1, OWL.onProperty, self._prop_uri(self.aliased_slot_name(slot))))

                            restr2 = BNode()
                            self.graph.add((restr2, RDF.type, OWL.Restriction))
                            self.graph.add((restr2, OWL.someValuesFrom, self._range_uri(slot)))
                            self.graph.add((restr2, OWL.onProperty, self._prop_uri(self.aliased_slot_name(slot))))

                            coll_bnode = BNode()
                            Collection(self.graph, coll_bnode, [restr1, restr2])
                            self.graph.add((slot_node, OWL.intersectionOf, coll_bnode))
                            self.graph.add((slot_node, RDF.type, OWL.Class))
                        else:
                            #    y         n      restriction(slot exactly 1 type)
                            self.graph.add((slot_node, RDF.type, OWL.Restriction))
                            self.graph.add((slot_node, OWL.qualifiedCardinality, Literal(1)))
                            self.graph.add((slot_node, OWL.onProperty, self._prop_uri(self.aliased_slot_name(slot))))
                            self.graph.add((slot_node, OWL.onClass, self._range_uri(slot)))
                    else:
                        if slot.multivalued:
                            #    n         y      restriction(slot only type)
                            self.graph.add((slot_node, RDF.type, OWL.Restriction))
                            self.graph.add((slot_node, OWL.allValuesFrom, self._range_uri(slot)))
                            self.graph.add((slot_node, OWL.onProperty, self._prop_uri(self.aliased_slot_name(slot))))
                        else:
                            #    n         n      intersectionOf(restriction(slot only type) restriction(slot max 1 type))
                            self.graph.add((slot_node, RDF.type, OWL.Restriction))
                            self.graph.add((slot_node, OWL.onClass, self._range_uri(slot)))
                            self.graph.add((slot_node, OWL.maxQualifiedCardinality, Literal(1)))
                            self.graph.add((slot_node, OWL.onProperty, self._prop_uri(self.aliased_slot_name(slot))))

        return True


    def visit_slot(self, slot_name: str, slot: SlotDefinition) -> None:
        """ Add a slot definition per slot

        @param slot_name:
        @param slot:
        @return:
        """
        # Note: We use the raw name in OWL and add a subProperty arc
        slot_uri = self._prop_uri(slot.name)
        self._add_element_properties(slot_uri, slot)
        # Inherited slots are object or data properties
        # All others are annotation properties
        self.graph.add((slot_uri, RDF.type, OWL.ObjectProperty if self.is_slot_object_property(slot) else OWL.AnnotationProperty))
        self.graph.add((slot_uri, RDF.type,
                        self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][camelcase('slot definition')]))
        self.graph.add((slot_uri, RDFS.range, self._range_uri(slot)))
        if slot.domain:
            self.graph.add((slot_uri, RDFS.domain, self._class_uri(slot.domain)))

        # Parent slots
        if slot.is_a:
            self.graph.add((slot_uri, RDFS.subPropertyOf, self._prop_uri(slot.is_a)))
        for mixin in slot.mixins:
            self.graph.add((slot_uri, RDFS.subPropertyOf, self._prop_uri(mixin)))
        if slot.name in self.synopsis.applytorefs:
            for appl in self.synopsis.applytorefs[slot.name].slotrefs:
                self.graph.add((slot_uri, RDFS.subClassOf, self._prop_uri(appl)))

    def visit_type(self, typ: TypeDefinition) -> None:
        type_uri = self._type_uri(typ.name)
        self.graph.add((type_uri, RDF.type, OWL.Class))
        self.graph.add((type_uri, RDF.type,
                        self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][camelcase('type definition')]))
        self._add_element_properties(type_uri, typ)
        if typ.typeof:
            self.graph.add((type_uri, RDFS.subClassOf, self._type_uri(typ.typeof)))
        else:
            restr = BNode()
            self.graph.add((restr, RDF.type, OWL.Restriction))
            self.graph.add((restr, OWL.qualifiedCardinality, Literal(1)))
            self.graph.add((restr, OWL.onProperty, self.top_value_uri))
            self.graph.add((restr, OWL.onDataRange, self.namespaces.uri_for(typ.uri)))
            self.graph.add((type_uri, RDFS.subClassOf, restr))

    def _add_element_properties(self, uri: URIRef, el: Element) -> None:
        for k, v in el.__dict__.items():
            if k in self.metamodel.schema.slots:
                defining_slot = self.metamodel.schema.slots[k]
                if v is not None and 'owl' in defining_slot.in_subset:
                    ve = v if isinstance(v, list) else [v]
                    for e in ve:
                        self.graph.add((uri,
                                        URIRef(self.metamodel.namespaces.uri_for(defining_slot.slot_uri)),
                                        Literal(e)))

    def _add_cardinality(self, subj: Union[BNode, URIRef], slot) -> None:
        """ Add cardinality restrictions to node """
        if slot.required:
            if slot.multivalued:
                self.graph.add((subj, OWL.minCardinality, Literal(1)))
            else:
                self.graph.add((subj, OWL.cardinality, Literal(1)))
        elif not slot.multivalued:
            self.graph.add((subj, OWL.maxCardinality, Literal(1)))

    def _range_uri(self, slot: SlotDefinition) -> URIRef:
        if slot.range in self.schema.types:
            return self._type_uri(cast(TypeDefinitionName, slot.range))
        else:
            return self._class_uri(cast(ClassDefinitionName, slot.range))

    def _class_uri(self, cn: ClassDefinitionName) -> URIRef:
        return self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][camelcase(cn)]

    def _prop_uri(self, pn: SlotDefinitionName) -> URIRef:
        return self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][underscore(pn)]

    def _type_uri(self, tn: TypeDefinitionName) -> URIRef:
        return self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][underscore(tn)]

    def _add_metamodel_class(self, cname: str) -> None:
        metac = self.metamodel.schema.classes[cname]
        metac_uri = self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME][camelcase(metac.name)]
        self.graph.add((metac_uri, RDF.type, OWL.Class))
        self._add_element_properties(metac_uri, metac)


    def is_slot_object_property(self, slot : SlotDefinition) -> bool:
        return True


@shared_arguments(OwlSchemaGenerator)
@click.command()
@click.option("-o", "--output", help="Output file name")
def cli(yamlfile, **kwargs):
    """ Generate an OWL representation of a biolink model """
    print(OwlSchemaGenerator(yamlfile, **kwargs).serialize(**kwargs))



import os
import re
import unittest

# This has to occur post ClickTestCase
from functools import reduce
from typing import List, Tuple
from urllib.parse import urljoin

import click

from biolinkml.generators.rdfgen import cli
from tests import source_yaml_path
from tests.test_scripts.clicktestcase import ClickTestCase
from tests import source_context_path


repl1: List[Tuple[str, str]] = [
    (r'(\s*):generation_date\s*".*"\^\^xsd:dateTime', r'\1:generation_date "2019-01-25 12:34"^^xsd:dateTime'),
    (r'(\s*):source_file_date\s*".*"\^\^xsd:dateTime', r'\1:source_file_date "2019-01-25 12:34"^^xsd:dateTime'),
    (r'(\s*):source_file_size\s*[0-9]+', r'\1:source_file_size 10000'),
]


def filtr(txt: str) -> str:
    return reduce(lambda s, expr: re.sub(expr[0], expr[1], s, flags=re.MULTILINE), repl1, txt)


class GenRDFTestCase(ClickTestCase):
    testdir = "genrdf"
    click_ep = cli
    prog_name = "gen-rdf"

    def test_help(self):
        self.do_test("--help", 'help', bypass_soft_compare=True)

    def test_meta(self):
        """ Test the RDF generator on the metamodel """
        meta_context_path = urljoin('file:', os.path.join(self.test_base_dir, 'gencontext', 'meta_context.jsonld'))
        meta_contextn_path = urljoin('file:', os.path.join(self.test_base_dir, 'gencontext', 'meta_contextn.jsonld'))
        self.maxDiff = None
        self.do_test(source_yaml_path + f" --context {meta_context_path}", 'meta.ttl', filtr=filtr,
                     comparator=ClickTestCase.rdf_comparator)
        self.do_test(source_yaml_path + f" --context {meta_contextn_path} --metauris", 'metan.ttl', filtr=filtr,
                     comparator=ClickTestCase.rdf_comparator)
        self.do_test(source_yaml_path + f' -f n3  --context {meta_context_path}', 'meta.n3', filtr=filtr,
                     comparator=ClickTestCase.rdf_comparator)
        self.do_test(source_yaml_path + f' -f xsv  --context {meta_context_path}', 'meta_error',
                     error=click.exceptions.BadParameter)

    def test_make_script(self):
        """ Test a relative file path in JSON """
        self.do_test(source_yaml_path + f" --context {source_context_path}",
                     'make_output.ttl', filtr=filtr, comparator=ClickTestCase.rdf_comparator)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from typing import List

from pyshex.shex_evaluator import EvaluationResult, ShExEvaluator
from rdflib import Graph

from biolinkml import LOCAL_METAMODEL_LDCONTEXT_FILE, METAMODEL_NAMESPACE, LOCAL_SHEXJ_FILE_NAME, \
    LOCAL_RDF_FILE_NAME, MODULE_DIR
from biolinkml.generators.jsonldgen import JSONLDGenerator
from biolinkml.generators.markdowngen import MarkdownGenerator
from biolinkml.generators.owlgen import OwlSchemaGenerator
from biolinkml.generators.rdfgen import RDFGenerator
from biolinkml.generators.shexgen import ShExGenerator
from tests import targetdir, DO_SHEX_VALIDATION, sourcedir
from tests.utils.generator_utils import GeneratorTestCase, BIOLINK_IMPORT_MAP
from tests.utils.metadata_filters import json_metadata_filter


class MetaModelTestCase(GeneratorTestCase):
    source_path = MODULE_DIR
    target_path = targetdir
    model_path = MODULE_DIR
    model_name = 'meta'
    importmap = BIOLINK_IMPORT_MAP

    @unittest.skipIf(True, 'Disable since docs generation is moved to GitHub Actions')
    def test_meta_markdown(self):
        """ Test the markdown generator for the biolink model """
        self.directory_generator('docs', MarkdownGenerator)

    @unittest.skipIf(False, "We still need to figure out what to do here")
    def test_meta_owl_schema(self):
        """ Test the owl schema generator for the biolink model """
        self.single_file_generator('owl', OwlSchemaGenerator, comparator=GeneratorTestCase.rdf_comparator)

    @staticmethod
    def _evaluate_shex_results(results: List[EvaluationResult]) -> bool:
        """
        Check the results of the ShEx evaluation
        :param results: evaluate output
        :return: success indicator
        """
        success = all(r.result for r in results)
        if not success:
            for r in results:
                if not r.result:
                    print(r.reason)
        return success

    def test_meta_shexc(self):
        """ Test the shex ShExC generation """
        self.single_file_generator('shex', ShExGenerator, format="shex")

    def test_meta_shecj(self):
        """ Test the shex ShExJ generation """
        self.single_file_generator('shexj', ShExGenerator, format="json")

    def test_meta_rdf(self):
        """ Test the rdf generator for the biolink model """
        # Make sure the ShEx is good
        self.single_file_generator('shexj', ShExGenerator, format="json")

        # Make a fresh copy of the RDF and validate it as well
        self.single_file_generator('ttl', RDFGenerator, serialize_args={"context": LOCAL_METAMODEL_LDCONTEXT_FILE},
                                   comparator=GeneratorTestCase.rdf_comparator)

        # Validate the RDF against the Biolink ShEx
        if DO_SHEX_VALIDATION:
            g = Graph()
            rdf_file = LOCAL_RDF_FILE_NAME
            g.load(rdf_file, format='turtle')
            focus = METAMODEL_NAMESPACE.metamodel
            start = METAMODEL_NAMESPACE.SchemaDefinition
            results = ShExEvaluator(g, LOCAL_SHEXJ_FILE_NAME, focus, start).evaluate(debug=False)
            self.assertTrue(self._evaluate_shex_results(results))
        else:
            print("*** RDF Model validation step was skipped. Set: tests.__init__.DO_SHEX_VALIDATION to run it")


if __name__ == '__main__':
    unittest.main()
